# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import unittest

import dpctl
import numpy as np

import numba_dppy as dppy

from . import _helper


@dppy.kernel
def reduction_kernel(A, R, stride):
    i = dppy.get_global_id(0)
    # sum two element
    R[i] = A[i] + A[i + stride]
    # store the sum to be used in nex iteration
    A[i] = R[i]


@unittest.skipUnless(_helper.has_gpu_queues(), "test only on GPU system")
class TestDPPYSumReduction(unittest.TestCase):
    def test_sum_reduction(self):
        # This test will only work for even case
        N = 1024
        self.assertTrue(N % 2 == 0)

        A = np.array(np.random.random(N), dtype=np.float32)
        A_copy = A.copy()
        # at max we will require half the size of A to store sum
        R = np.array(np.random.random(math.ceil(N / 2)), dtype=np.float32)

        device = dpctl.SyclDevice("opencl:gpu")
        with dpctl.device_context(device):
            total = N

            while total > 1:
                # call kernel
                global_size = total // 2
                reduction_kernel[global_size, dppy.DEFAULT_LOCAL_SIZE](
                    A, R, global_size
                )
                total = total // 2

            result = A_copy.sum()
            max_abs_err = result - R[0]
            self.assertTrue(max_abs_err < 1e-4)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Various utility functions and classes to aid LLVM IR building.

"""
from numba_dppy.utils.array_utils import (
    as_usm_obj,
    copy_from_numpy_to_usm_obj,
    copy_to_numpy_from_usm_obj,
    has_usm_memory,
)
from numba_dppy.utils.constants import address_space, calling_conv
from numba_dppy.utils.llvm_codegen_helpers import (
    LLVMTypes,
    create_null_ptr,
    get_llvm_ptr_type,
    get_llvm_type,
    get_one,
    get_zero,
)
from numba_dppy.utils.misc import assert_no_return
from numba_dppy.utils.type_conversion_fns import npytypes_array_to_dppy_array

__all__ = [
    "LLVMTypes",
    "get_llvm_type",
    "get_llvm_ptr_type",
    "create_null_ptr",
    "get_zero",
    "get_one",
    "npytypes_array_to_dppy_array",
    "address_space",
    "calling_conv",
    "has_usm_memory",
    "as_usm_obj",
    "copy_from_numpy_to_usm_obj",
    "copy_to_numpy_from_usm_obj",
    "assert_no_return",
]



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import llvmlite.binding as ll
from numba.np.ufunc.decorators import Vectorize

from numba_dppy.vectorizers import DPPYVectorize


def init_jit():
    from numba_dppy.dispatcher import DPPYDispatcher

    return DPPYDispatcher


def load_dpctl_sycl_interface():
    """Permanently loads the ``DPCTLSyclInterface`` library provided by dpctl.

    The ``DPCTLSyclInterface`` library provides C wrappers over SYCL functions
    that are directly invoked from the LLVM modules generated by numba-dppy.
    We load the library once at the time of initialization using llvmlite's
    load_library_permanently function.

    Raises:
        ImportError: If the ``DPCTLSyclInterface`` library could not be loaded.
    """
    import glob
    import platform as plt

    import dpctl

    platform = plt.system()
    if platform == "Windows":
        paths = glob.glob(
            os.path.join(os.path.dirname(dpctl.__file__), "*DPCTLSyclInterface.dll")
        )
    else:
        paths = glob.glob(
            os.path.join(os.path.dirname(dpctl.__file__), "*DPCTLSyclInterface.so")
        )

    if len(paths) == 1:
        ll.load_library_permanently(paths[0])
    else:
        raise ImportError

    def init_dppy_vectorize():
        return DPPYVectorize

    Vectorize.target_registry.ondemand["dppy"] = init_dppy_vectorize



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dpctl
from numba.core import sigutils, types

from numba_dppy.utils import assert_no_return, npytypes_array_to_dppy_array

from .compiler import (
    JitDPPYKernel,
    compile_dppy_func,
    compile_dppy_func_template,
    get_ordered_arg_access_types,
)


def kernel(signature=None, access_types=None, debug=None):
    """JIT compile a python function conforming using the DPPY backend.

    A kernel is equvalent to an OpenCL kernel function, and has the
    same restrictions as definined by SPIR_KERNEL calling convention.
    """
    if signature is None:
        return autojit(debug=debug, access_types=access_types)
    elif not sigutils.is_signature(signature):
        func = signature
        return autojit(debug=debug, access_types=access_types)(func)
    else:
        return _kernel_jit(signature, debug, access_types)


def autojit(debug=None, access_types=None):
    def _kernel_autojit(pyfunc):
        ordered_arg_access_types = get_ordered_arg_access_types(pyfunc, access_types)
        return JitDPPYKernel(pyfunc, debug, ordered_arg_access_types)

    return _kernel_autojit


def _kernel_jit(signature, debug, access_types):
    argtypes, rettype = sigutils.normalize_signature(signature)
    argtypes = tuple(
        [
            npytypes_array_to_dppy_array(ty)
            if isinstance(ty, types.npytypes.Array)
            else ty
            for ty in argtypes
        ]
    )

    # Raises TypeError when users return anything inside @dppy.kernel.
    assert_no_return(rettype)

    def _wrapped(pyfunc):
        current_queue = dpctl.get_current_queue()
        ordered_arg_access_types = get_ordered_arg_access_types(pyfunc, access_types)
        # We create an instance of JitDPPYKernel to make sure at call time
        # we are going through the caching mechanism.
        dppy_kernel = JitDPPYKernel(pyfunc, debug, ordered_arg_access_types)
        # This will make sure we are compiling eagerly.
        dppy_kernel.specialize(argtypes, current_queue)
        return dppy_kernel

    return _wrapped


def func(signature=None, debug=None):
    if signature is None:
        return _func_autojit_wrapper(debug=debug)
    elif not sigutils.is_signature(signature):
        func = signature
        return _func_autojit(func, debug=debug)
    else:
        return _func_jit(signature, debug=debug)


def _func_jit(signature, debug=None):
    argtypes, restype = sigutils.normalize_signature(signature)
    argtypes = tuple(
        [
            npytypes_array_to_dppy_array(ty)
            if isinstance(ty, types.npytypes.Array)
            else ty
            for ty in argtypes
        ]
    )

    def _wrapped(pyfunc):
        return compile_dppy_func(pyfunc, restype, argtypes, debug=debug)

    return _wrapped


def _func_autojit_wrapper(debug=None):
    def _func_autojit(pyfunc, debug=debug):
        return compile_dppy_func_template(pyfunc, debug=debug)

    return _func_autojit


def _func_autojit(pyfunc, debug=None):
    return compile_dppy_func_template(pyfunc, debug=debug)



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numba import types
from numba.core.extending import overload, register_jitable
from numba.core.typing import signature

import numba_dppy
import numba_dppy.dpnp_glue as dpnp_lowering
import numba_dppy.dpnp_glue.dpnpimpl as dpnp_ext
from numba_dppy import dpctl_functions

from . import stubs


@overload(stubs.dpnp.eig)
def dpnp_eig_impl(a):
    name = "eig"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels.cpp#L180

    Function declaration:
    void dpnp_eig_c(const void* array_in, void* result1, void* result2, size_t size)

    """
    sig = signature(ret_type, types.voidptr, types.voidptr, types.voidptr, types.intp)
    dpnp_eig = dpnp_ext.dpnp_func("dpnp_" + name, [a.dtype.name, "NONE"], sig)

    res_dtype = np.float64
    if a.dtype == types.float32:
        res_dtype = np.float32
    PRINT_DEBUG = dpnp_lowering.DEBUG

    def dpnp_impl(a):
        n = a.shape[-1]
        if a.shape[-2] != n:
            msg = "Last 2 dimensions of the array must be square."
            raise ValueError(msg)

        dpnp_ext._check_finite_matrix(a)

        wr = np.empty(n, dtype=res_dtype)
        vr = np.empty((n, n), dtype=res_dtype)

        if n == 0:
            return (wr, vr)

        sycl_queue = dpctl_functions.get_current_queue()
        a_usm = dpctl_functions.malloc_shared(a.size * a.itemsize, sycl_queue)
        event = dpctl_functions.queue_memcpy(
            sycl_queue, a_usm, a.ctypes, a.size * a.itemsize
        )
        dpctl_functions.event_wait(event)
        dpctl_functions.event_delete(event)

        wr_usm = dpctl_functions.malloc_shared(wr.size * wr.itemsize, sycl_queue)
        vr_usm = dpctl_functions.malloc_shared(vr.size * vr.itemsize, sycl_queue)

        dpnp_eig(a_usm, wr_usm, vr_usm, n)

        event = dpctl_functions.queue_memcpy(
            sycl_queue, wr.ctypes, wr_usm, wr.size * wr.itemsize
        )
        dpctl_functions.event_wait(event)
        dpctl_functions.event_delete(event)
        event = dpctl_functions.queue_memcpy(
            sycl_queue, vr.ctypes, vr_usm, vr.size * vr.itemsize
        )
        dpctl_functions.event_wait(event)
        dpctl_functions.event_delete(event)

        dpctl_functions.free_with_queue(a_usm, sycl_queue)
        dpctl_functions.free_with_queue(wr_usm, sycl_queue)
        dpctl_functions.free_with_queue(vr_usm, sycl_queue)

        dpnp_ext._dummy_liveness_func([wr.size, vr.size])

        if PRINT_DEBUG:
            print("dpnp implementation")
        return (wr, vr)

    return dpnp_impl


@register_jitable
def common_matmul_impl(dpnp_func, a, b, out, m, n, k, print_debug):
    sycl_queue = dpctl_functions.get_current_queue()

    a_usm = dpctl_functions.malloc_shared(a.size * a.itemsize, sycl_queue)
    dpctl_functions.queue_memcpy(sycl_queue, a_usm, a.ctypes, a.size * a.itemsize)

    b_usm = dpctl_functions.malloc_shared(b.size * b.itemsize, sycl_queue)
    event = dpctl_functions.queue_memcpy(
        sycl_queue, b_usm, b.ctypes, b.size * b.itemsize
    )
    dpctl_functions.event_wait(event)
    dpctl_functions.event_delete(event)

    out_usm = dpctl_functions.malloc_shared(out.size * out.itemsize, sycl_queue)

    a_m = np.array((m, k))
    b_m = np.array((k, n))
    out_m = np.array((m, n))

    strides = np.array(1)

    result_out = out_usm
    result_size = out.size
    result_ndim = 2
    result_shape = out_m.ctypes
    result_strides = strides.ctypes

    input1_in = a_usm
    input1_size = a.size
    input1_ndim = 2
    input1_shape = a_m.ctypes
    input1_strides = strides.ctypes

    input2_in = b_usm
    input2_size = b.size
    input2_ndim = 2
    input2_shape = b_m.ctypes
    input2_strides = strides.ctypes

    dpnp_func(
        result_out,
        result_size,
        result_ndim,
        result_shape,
        result_strides,
        input1_in,
        input1_size,
        input1_ndim,
        input1_shape,
        input1_strides,
        input2_in,
        input2_size,
        input2_ndim,
        input2_shape,
        input2_strides,
    )

    event = dpctl_functions.queue_memcpy(
        sycl_queue, out.ctypes, out_usm, out.size * out.itemsize
    )
    dpctl_functions.event_wait(event)
    dpctl_functions.event_delete(event)

    dpctl_functions.free_with_queue(a_usm, sycl_queue)
    dpctl_functions.free_with_queue(b_usm, sycl_queue)
    dpctl_functions.free_with_queue(out_usm, sycl_queue)

    dpnp_ext._dummy_liveness_func([a.size, b.size, out.size])
    dpnp_ext._dummy_liveness_func([a_m.size, b_m.size, out_m.size])

    if print_debug:
        print("dpnp implementation")


@register_jitable
def common_dot_impl(dpnp_func, a, b, out, m, print_debug):
    sycl_queue = dpctl_functions.get_current_queue()
    a_usm = dpctl_functions.malloc_shared(a.size * a.itemsize, sycl_queue)
    event = dpctl_functions.queue_memcpy(
        sycl_queue, a_usm, a.ctypes, a.size * a.itemsize
    )
    dpctl_functions.event_wait(event)
    dpctl_functions.event_delete(event)

    b_usm = dpctl_functions.malloc_shared(b.size * b.itemsize, sycl_queue)
    event = dpctl_functions.queue_memcpy(
        sycl_queue, b_usm, b.ctypes, b.size * b.itemsize
    )
    dpctl_functions.event_wait(event)
    dpctl_functions.event_delete(event)

    out_usm = dpctl_functions.malloc_shared(out.size * out.itemsize, sycl_queue)

    strides = np.array(1)

    result_out = out_usm
    result_size = out.size
    result_ndim = out.ndim
    result_shape = out.shapeptr
    result_strides = strides.ctypes

    input1_in = a_usm
    input1_size = a.size
    input1_ndim = a.ndim
    input1_shape = a.shapeptr
    input1_strides = strides.ctypes

    input2_in = b_usm
    input2_size = b.size
    input2_ndim = b.ndim
    input2_shape = b.shapeptr
    input2_strides = strides.ctypes

    dpnp_func(
        result_out,
        result_size,
        result_ndim,
        result_shape,
        result_strides,
        input1_in,
        input1_size,
        input1_ndim,
        input1_shape,
        input1_strides,
        input2_in,
        input2_size,
        input2_ndim,
        input2_shape,
        input2_strides,
    )

    event = dpctl_functions.queue_memcpy(
        sycl_queue, out.ctypes, out_usm, out.size * out.itemsize
    )
    dpctl_functions.event_wait(event)
    dpctl_functions.event_delete(event)

    dpctl_functions.free_with_queue(a_usm, sycl_queue)
    dpctl_functions.free_with_queue(b_usm, sycl_queue)
    dpctl_functions.free_with_queue(out_usm, sycl_queue)

    dpnp_ext._dummy_liveness_func([a.size, b.size, out.size])

    if print_debug:
        print("dpnp implementation")


def get_res_dtype(a, b):
    res_dtype = np.float64
    if a.dtype == types.int32 and b.dtype == types.int32:
        res_dtype = np.int32
    elif a.dtype == types.int32 and b.dtype == types.int64:
        res_dtype = np.int64
    elif a.dtype == types.int32 and b.dtype == types.float32:
        res_dtype = np.float64
    elif a.dtype == types.int32 and b.dtype == types.float64:
        res_dtype = np.float64
    elif a.dtype == types.int64 and b.dtype == types.int32:
        res_dtype = np.int64
    elif a.dtype == types.int64 and b.dtype == types.int64:
        res_dtype = np.int64
    elif a.dtype == types.int64 and b.dtype == types.float32:
        res_dtype = np.float64
    elif a.dtype == types.int64 and b.dtype == types.float64:
        res_dtype = np.float64
    elif a.dtype == types.float32 and b.dtype == types.int32:
        res_dtype = np.float64
    elif a.dtype == types.float32 and b.dtype == types.int64:
        res_dtype = np.float64
    elif a.dtype == types.float32 and b.dtype == types.float32:
        res_dtype = np.float32
    elif a.dtype == types.float32 and b.dtype == types.float64:
        res_dtype = np.float64
    elif a.dtype == types.float64 and b.dtype == types.int32:
        res_dtype = np.float64
    elif a.dtype == types.float64 and b.dtype == types.int64:
        res_dtype = np.float64
    elif a.dtype == types.float64 and b.dtype == types.float32:
        res_dtype = np.float32
    elif a.dtype == types.float64 and b.dtype == types.float64:
        res_dtype = np.float64

    return res_dtype


@overload(stubs.dpnp.matmul)
@overload(stubs.dpnp.dot)
def dpnp_dot_impl(a, b):
    dpnp_lowering.ensure_dpnp("dot")

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blame/67a101c90cf253cfe9b9ba80ac397811ce94edee/dpnp/backend/kernels/dpnp_krnl_common.cpp#L322

    Function declaration:
    void dpnp_matmul_c(void* result_out,
                    const size_t result_size,
                    const size_t result_ndim,
                    const size_t* result_shape,
                    const size_t* result_strides,
                    const void* input1_in,
                    const size_t input1_size,
                    const size_t input1_ndim,
                    const size_t* input1_shape,
                    const size_t* input1_strides,
                    const void* input2_in,
                    const size_t input2_size,
                    const size_t input2_ndim,
                    const size_t* input2_shape,
                    const size_t* input2_strides)
    """
    sig = signature(
        ret_type,
        types.voidptr,
        types.intp,
        types.intp,
        types.voidptr,
        types.voidptr,
        types.voidptr,
        types.intp,
        types.intp,
        types.voidptr,
        types.voidptr,
        types.voidptr,
        types.intp,
        types.intp,
        types.voidptr,
        types.voidptr,
    )

    res_dtype = get_res_dtype(a, b)

    PRINT_DEBUG = dpnp_lowering.DEBUG

    ndims = [a.ndim, b.ndim]
    if ndims == [2, 2]:
        dpnp_func = dpnp_ext.dpnp_func("dpnp_matmul", [a.dtype.name, "NONE"], sig)

        def dot_2_mm(a, b):
            m, k = a.shape
            _k, n = b.shape

            if _k != k:
                raise ValueError("Incompatible array sizes for np.dot(a, b)")

            out = np.empty((m, n), dtype=res_dtype)
            common_matmul_impl(dpnp_func, a, b, out, m, n, k, PRINT_DEBUG)

            return out

        return dot_2_mm
    elif ndims == [2, 1]:
        dpnp_func = dpnp_ext.dpnp_func("dpnp_matmul", [a.dtype.name, "NONE"], sig)

        def dot_2_mv(a, b):
            m, k = a.shape
            (_n,) = b.shape
            n = 1

            if _n != k:
                raise ValueError("Incompatible array sizes for np.dot(a, b)")

            out = np.empty((m,), dtype=res_dtype)
            common_matmul_impl(dpnp_func, a, b, out, m, n, k, PRINT_DEBUG)

            return out

        return dot_2_mv
    elif ndims == [1, 2]:
        dpnp_func = dpnp_ext.dpnp_func("dpnp_matmul", [a.dtype.name, "NONE"], sig)

        def dot_2_vm(a, b):
            (m,) = a.shape
            k, n = b.shape

            if m != k:
                raise ValueError("Incompatible array sizes for np.dot(a, b)")

            out = np.empty((n,), dtype=res_dtype)
            common_matmul_impl(dpnp_func, a, b, out, m, n, k, PRINT_DEBUG)

            return out

        return dot_2_vm
    elif ndims == [1, 1]:
        """
        dpnp source:
        https://github.com/IntelPython/dpnp/blob/67a101c90cf253cfe9b9ba80ac397811ce94edee/dpnp/backend/kernels/dpnp_krnl_common.cpp#L79

        Function declaration:
        void dpnp_dot_c(void* result_out,
                        const size_t result_size,
                        const size_t result_ndim,
                        const size_t* result_shape,
                        const size_t* result_strides,
                        const void* input1_in,
                        const size_t input1_size,
                        const size_t input1_ndim,
                        const size_t* input1_shape,
                        const size_t* input1_strides,
                        const void* input2_in,
                        const size_t input2_size,
                        const size_t input2_ndim,
                        const size_t* input2_shape,
                        const size_t* input2_strides)
        """
        sig = signature(
            ret_type,
            types.voidptr,
            types.intp,
            types.intp,
            types.voidptr,
            types.voidptr,
            types.voidptr,
            types.intp,
            types.intp,
            types.voidptr,
            types.voidptr,
            types.voidptr,
            types.intp,
            types.intp,
            types.voidptr,
            types.voidptr,
        )
        dpnp_func = dpnp_ext.dpnp_func("dpnp_dot", [a.dtype.name, "NONE"], sig)

        def dot_2_vv(a, b):

            (m,) = a.shape
            (n,) = b.shape

            if m != n:
                raise ValueError("Incompatible array sizes for np.dot(a, b)")

            out = np.empty(1, dtype=res_dtype)
            common_dot_impl(dpnp_func, a, b, out, m, PRINT_DEBUG)

            return out[0]

        return dot_2_vv
    else:
        assert 0


@overload(stubs.dpnp.multi_dot)
def dpnp_multi_dot_impl(arrays):
    dpnp_lowering.ensure_dpnp("multi_dot")

    print_debug = dpnp_lowering.DEBUG
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels.cpp#L118

    Function declaration:
    void dpnp_dot_c(void* array1_in, void* array2_in, void* result1, size_t size)

    """

    def dpnp_impl(arrays):
        n = len(arrays)
        result = arrays[0]

        for idx in range(1, n):
            result = numba_dppy.dpnp.dot(result, arrays[idx])

        if print_debug:
            print("dpnp implementation")
        return result

    return dpnp_impl


@overload(stubs.dpnp.vdot)
def dpnp_vdot_impl(a, b):
    dpnp_lowering.ensure_dpnp("vdot")

    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels.cpp#L118

    Function declaration:
    void dpnp_dot_c(void* array1_in, void* array2_in, void* result1, size_t size)

    """

    def dpnp_impl(a, b):
        return numba_dppy.dpnp.dot(np.ravel(a), np.ravel(b))

    return dpnp_impl


@overload(stubs.dpnp.matrix_power)
def dpnp_matrix_power_impl(a, n):
    dpnp_lowering.ensure_dpnp("matrix_power")

    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels.cpp#L42

    Function declaration:
    void dpnp_matmul_c(void* array1_in, void* array2_in, void* result1, size_t size_m,
                       size_t size_n, size_t size_k)
    """

    PRINT_DEBUG = dpnp_lowering.DEBUG

    def dpnp_impl(a, n):
        if n < 0:
            raise ValueError("n < 0 is not supported for np.linalg.matrix_power(a, n)")

        if n == 0:
            if PRINT_DEBUG:
                print("dpnp implementation")
            return np.identity(a.shape[0], a.dtype)

        result = a
        for idx in range(0, n - 1):
            result = numba_dppy.dpnp.matmul(result, a)
        return result

    return dpnp_impl


@overload(stubs.dpnp.cholesky)
def dpnp_cholesky_impl(a):
    name = "cholesky"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_linalg.cpp#L40

    Function declaration:
    void custom_cholesky_c(void* array1_in, void* result1, size_t* shape)

    """
    sig = signature(ret_type, types.voidptr, types.voidptr, types.voidptr)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, [a.dtype.name, "NONE"], sig)
    PRINT_DEBUG = dpnp_lowering.DEBUG

    def dpnp_impl(a):
        n = a.shape[-1]
        if a.shape[-2] != n:
            raise ValueError("Input array must be square.")

        out = a.copy()

        if n == 0:
            return out

        sycl_queue = dpctl_functions.get_current_queue()
        a_usm = dpctl_functions.malloc_shared(a.size * a.itemsize, sycl_queue)
        event = dpctl_functions.queue_memcpy(
            sycl_queue, a_usm, a.ctypes, a.size * a.itemsize
        )
        dpctl_functions.event_wait(event)
        dpctl_functions.event_delete(event)

        out_usm = dpctl_functions.malloc_shared(out.size * out.itemsize, sycl_queue)

        dpnp_func(a_usm, out_usm, a.shapeptr)

        event = dpctl_functions.queue_memcpy(
            sycl_queue, out.ctypes, out_usm, out.size * out.itemsize
        )
        dpctl_functions.event_wait(event)
        dpctl_functions.event_delete(event)

        dpctl_functions.free_with_queue(a_usm, sycl_queue)
        dpctl_functions.free_with_queue(out_usm, sycl_queue)

        dpnp_ext._dummy_liveness_func([out.size, a.size])

        if PRINT_DEBUG:
            print("dpnp implementation")
        return out

    return dpnp_impl


@overload(stubs.dpnp.det)
def dpnp_det_impl(a):
    name = "det"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_linalg.cpp#L83

    Function declaration:
    void custom_det_c(void* array1_in, void* result1, size_t* shape, size_t ndim)
    """
    sig = signature(ret_type, types.voidptr, types.voidptr, types.voidptr, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, [a.dtype.name, "NONE"], sig)
    PRINT_DEBUG = dpnp_lowering.DEBUG

    def dpnp_impl(a):
        n = a.shape[-1]
        if a.shape[-2] != n:
            raise ValueError("Input array must be square.")

        dpnp_ext._check_finite_matrix(a)

        if a.ndim == 2:
            out = np.empty((1,), dtype=a.dtype)
            out[0] = -4
        else:
            out = np.empty(a.shape[:-2], dtype=a.dtype)

        sycl_queue = dpctl_functions.get_current_queue()
        a_usm = dpctl_functions.malloc_shared(a.size * a.itemsize, sycl_queue)
        event = dpctl_functions.queue_memcpy(
            sycl_queue, a_usm, a.ctypes, a.size * a.itemsize
        )
        dpctl_functions.event_wait(event)
        dpctl_functions.event_delete(event)

        out_usm = dpctl_functions.malloc_shared(out.size * out.itemsize, sycl_queue)

        dpnp_func(a_usm, out_usm, a.shapeptr, a.ndim)

        event = dpctl_functions.queue_memcpy(
            sycl_queue, out.ctypes, out_usm, out.size * out.itemsize
        )
        dpctl_functions.event_wait(event)
        dpctl_functions.event_delete(event)

        dpctl_functions.free_with_queue(a_usm, sycl_queue)
        dpctl_functions.free_with_queue(out_usm, sycl_queue)

        dpnp_ext._dummy_liveness_func([out.size, a.size])

        if PRINT_DEBUG:
            print("dpnp implementation")
        if a.ndim == 2:
            return out[0]
        else:
            return out

    return dpnp_impl


@overload(stubs.dpnp.matrix_rank)
def dpnp_matrix_rank_impl(M, tol=None, hermitian=False):
    name = "matrix_rank"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_linalg.cpp#L186

    Function declaration:
    void custom_matrix_rank_c(void* array1_in, void* result1, size_t* shape, size_t ndim)
    """
    sig = signature(ret_type, types.voidptr, types.voidptr, types.voidptr, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, [M.dtype.name, "NONE"], sig)
    PRINT_DEBUG = dpnp_lowering.DEBUG

    def dpnp_impl(M, tol=None, hermitian=False):
        if tol is not None:
            raise ValueError("tol is not supported for np.linalg.matrix_rank(M)")
        if hermitian:
            raise ValueError("hermitian is not supported for np.linalg.matrix_rank(M)")

        if M.ndim > 2:
            raise ValueError(
                "np.linalg.matrix_rank(M) is only supported on 1 or 2-d arrays"
            )

        out = np.empty(1, dtype=M.dtype)

        sycl_queue = dpctl_functions.get_current_queue()
        M_usm = dpctl_functions.malloc_shared(M.size * M.itemsize, sycl_queue)
        event = dpctl_functions.queue_memcpy(
            sycl_queue, M_usm, M.ctypes, M.size * M.itemsize
        )
        dpctl_functions.event_wait(event)
        dpctl_functions.event_delete(event)

        out_usm = dpctl_functions.malloc_shared(out.size * out.itemsize, sycl_queue)

        dpnp_func(M_usm, out_usm, M.shapeptr, M.ndim)

        event = dpctl_functions.queue_memcpy(
            sycl_queue, out.ctypes, out_usm, out.size * out.itemsize
        )
        dpctl_functions.event_wait(event)
        dpctl_functions.event_delete(event)

        dpctl_functions.free_with_queue(M_usm, sycl_queue)
        dpctl_functions.free_with_queue(out_usm, sycl_queue)

        dpnp_ext._dummy_liveness_func([out.size, M.size])

        if PRINT_DEBUG:
            print("dpnp implementation")
        return out[0]

    return dpnp_impl


@overload(stubs.dpnp.eigvals)
def dpnp_eigvals_impl(a):
    dpnp_lowering.ensure_dpnp("eigvals")

    def dpnp_impl(a):
        eigval, eigvec = numba_dppy.dpnp.eig(a)
        return eigval

    return dpnp_impl



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numba.core import ir, types
from numba.core.compiler_machinery import FunctionPass, register_pass
from numba.core.ir_utils import (
    find_topo_order,
    mk_unique_var,
    remove_dead,
    simplify_CFG,
)

import numba_dppy

rewrite_function_name_map = {
    # numpy
    "all": (["numpy"], "all"),
    "amax": (["numpy"], "amax"),
    "amin": (["numpy"], "amin"),
    "argmax": (["numpy"], "argmax"),
    "argmin": (["numpy"], "argmin"),
    "argsort": (["numpy"], "argsort"),
    "cov": (["numpy"], "cov"),
    "diagonal": (["numpy"], "diagonal"),
    "max": (["numpy"], "max"),
    "mean": (["numpy"], "mean"),
    "median": (["numpy"], "median"),
    "min": (["numpy"], "min"),
    "partition": (["numpy"], "partition"),
    "repeat": (["numpy"], "repeat"),
    "trace": (["numpy"], "trace"),
    "vdot": (["numpy"], "vdot"),
    # random
    "beta": (["random"], "beta"),
    "binomial": (["random"], "binomial"),
    "chisquare": (["random"], "chisquare"),
    "exponential": (["random"], "exponential"),
    "gamma": (["random"], "gamma"),
    "geometric": (["random"], "geometric"),
    "gumbel": (["random"], "gumbel"),
    "hypergeometric": (["random"], "hypergeometric"),
    "laplace": (["random"], "laplace"),
    "lognormal": (["random"], "lognormal"),
    "multinomial": (["random"], "multinomial"),
    "multivariate_normal": (["random"], "multivariate_normal"),
    "negative_binomial": (["random"], "negative_binomial"),
    "normal": (["random"], "normal"),
    "poisson": (["random"], "poisson"),
    "rand": (["random"], "rand"),
    "randint": (["random"], "randint"),
    "random_integers": (["random"], "random_integers"),
    "random_sample": (["random"], "random_sample"),
    "random": (["random"], "random"),
    "ranf": (["random"], "ranf"),
    "rayleigh": (["random"], "rayleigh"),
    "sample": (["random"], "sample"),
    "standard_cauchy": (["random"], "standard_cauchy"),
    "standard_exponential": (["random"], "standard_exponential"),
    "standard_gamma": (["random"], "standard_gamma"),
    "standard_normal": (["random"], "standard_normal"),
    "uniform": (["random"], "uniform"),
    "weibull": (["random"], "weibull"),
    # linalg
    "cholesky": (["linalg"], "cholesky"),
    "det": (["linalg"], "det"),
    "dot": (["numpy"], "dot"),
    "eig": (["linalg"], "eig"),
    "eigvals": (["linalg"], "eigvals"),
    "matmul": (["numpy"], "matmul"),
    "matrix_power": (["linalg"], "matrix_power"),
    "matrix_rank": (["linalg"], "matrix_rank"),
    "multi_dot": (["linalg"], "multi_dot"),
    # transcendentals
    "nanprod": (["numpy"], "nanprod"),
    "nansum": (["numpy"], "nansum"),
    "prod": (["numpy"], "prod"),
    "sum": (["numpy"], "sum"),
    # array creations
    "full": (["numpy"], "full"),
    "ones_like": (["numpy"], "ones_like"),
    "zeros_like": (["numpy"], "zeros_like"),
    "full_like": (["numpy"], "full_like"),
    # array ops
    "copy": (["numpy"], "copy"),
    "cumsum": (["numpy"], "cumsum"),
    "cumprod": (["numpy"], "cumprod"),
    "sort": (["numpy"], "sort"),
    "take": (["numpy"], "take"),
}


class RewriteNumPyOverloadedFunctions(object):
    def __init__(self, state, rewrite_function_name_map=rewrite_function_name_map):
        self.state = state
        self.function_name_map = rewrite_function_name_map

    def run(self):
        """
        This function rewrites the name of NumPy functions that exist in self.function_name_map
        e.g np.sum(a) would produce the following:

        np.sum() --> numba_dppy.dpnp.sum()

        ---------------------------------------------------------------------------------------
        Numba IR Before Rewrite:
        ---------------------------------------------------------------------------------------

            $2load_global.0 = global(np: <module 'numpy' from 'numpy/__init__.py'>) ['$2load_global.0']
            $4load_method.1 = getattr(value=$2load_global.0, attr=sum) ['$2load_global.0', '$4load_method.1']
            $8call_method.3 = call $4load_method.1(a, func=$4load_method.1, args=[Var(a, test_rewrite.py:7)],
                                                   kws=(), vararg=None) ['$4load_method.1', '$8call_method.3', 'a']

        ---------------------------------------------------------------------------------------
        Numba IR After Rewrite:
        ---------------------------------------------------------------------------------------

            $dppy_replaced_var.0 = global(numba_dppy: <module 'numba_dppy' from 'numba_dppy/__init__.py'>) ['$dppy_replaced_var.0']
            $dpnp_var.1 = getattr(value=$dppy_replaced_var.0, attr=dpnp) ['$dpnp_var.1', '$dppy_replaced_var.0']
            $4load_method.1 = getattr(value=$dpnp_var.1, attr=sum) ['$4load_method.1', '$dpnp_var.1']
            $8call_method.3 = call $4load_method.1(a, func=$4load_method.1, args=[Var(a, test_rewrite.py:7)],
                                                   kws=(), vararg=None) ['$4load_method.1', '$8call_method.3', 'a']

        ---------------------------------------------------------------------------------------
        """
        func_ir = self.state.func_ir
        blocks = func_ir.blocks
        topo_order = find_topo_order(blocks)
        replaced = False

        for label in topo_order:
            block = blocks[label]
            saved_arr_arg = {}
            new_body = []
            for stmt in block.body:
                if isinstance(stmt, ir.Assign) and isinstance(stmt.value, ir.Expr):
                    lhs = stmt.target.name
                    rhs = stmt.value
                    # replace np.FOO with name from self.function_name_map["FOO"]
                    # e.g. np.sum will be replaced with numba_dppy.dpnp.sum
                    if rhs.op == "getattr" and rhs.attr in self.function_name_map:
                        module_node = block.find_variable_assignment(
                            rhs.value.name
                        ).value
                        if (
                            isinstance(module_node, ir.Global)
                            and module_node.value.__name__
                            in self.function_name_map[rhs.attr][0]
                        ) or (
                            isinstance(module_node, ir.Expr)
                            and module_node.attr in self.function_name_map[rhs.attr][0]
                        ):
                            rhs = stmt.value
                            rhs.attr = self.function_name_map[rhs.attr][1]

                            global_module = rhs.value
                            saved_arr_arg[lhs] = global_module

                            scope = global_module.scope
                            loc = global_module.loc

                            g_dppy_var = ir.Var(
                                scope, mk_unique_var("$2load_global"), loc
                            )
                            # We are trying to rename np.function_name/np.linalg.function_name with
                            # numba_dppy.dpnp.function_name.
                            # Hence, we need to have a global variable representing module numba_dppy.
                            # Next, we add attribute dpnp to global module numba_dppy to
                            # represent numba_dppy.dpnp.
                            g_dppy = ir.Global("numba_dppy", numba_dppy, loc)
                            g_dppy_assign = ir.Assign(g_dppy, g_dppy_var, loc)

                            dpnp_var = ir.Var(scope, mk_unique_var("$4load_attr"), loc)
                            getattr_dpnp = ir.Expr.getattr(g_dppy_var, "dpnp", loc)
                            dpnp_assign = ir.Assign(getattr_dpnp, dpnp_var, loc)

                            rhs.value = dpnp_var
                            new_body.append(g_dppy_assign)
                            new_body.append(dpnp_assign)
                            func_ir._definitions[dpnp_var.name] = [getattr_dpnp]
                            func_ir._definitions[g_dppy_var.name] = [g_dppy]
                            replaced = True

                new_body.append(stmt)
            block.body = new_body
            return replaced


@register_pass(mutates_CFG=True, analysis_only=False)
class DPPYRewriteOverloadedNumPyFunctions(FunctionPass):
    _name = "dppy_rewrite_overloaded_functions_pass"

    def __init__(self):
        FunctionPass.__init__(self)

        import numba_dppy.dpnp_glue.dpnp_array_creations_impl
        import numba_dppy.dpnp_glue.dpnp_array_ops_impl
        import numba_dppy.dpnp_glue.dpnp_indexing
        import numba_dppy.dpnp_glue.dpnp_linalgimpl
        import numba_dppy.dpnp_glue.dpnp_logic
        import numba_dppy.dpnp_glue.dpnp_manipulation
        import numba_dppy.dpnp_glue.dpnp_randomimpl
        import numba_dppy.dpnp_glue.dpnp_sort_search_countimpl
        import numba_dppy.dpnp_glue.dpnp_statisticsimpl
        import numba_dppy.dpnp_glue.dpnp_transcendentalsimpl
        import numba_dppy.dpnp_glue.dpnpdecl
        import numba_dppy.dpnp_glue.dpnpimpl

    def run_pass(self, state):
        rewrite_function_name_pass = RewriteNumPyOverloadedFunctions(
            state, rewrite_function_name_map
        )

        mutated = rewrite_function_name_pass.run()

        if mutated:
            remove_dead(state.func_ir.blocks, state.func_ir.arg_names, state.func_ir)
        state.func_ir.blocks = simplify_CFG(state.func_ir.blocks)

        return mutated


def get_dpnp_func_typ(func):
    from numba.core.typing.templates import builtin_registry

    for (k, v) in builtin_registry.globals:
        if k == func:
            return v
    raise RuntimeError("type for func ", func, " not found")


class RewriteNdarrayFunctions(object):
    def __init__(self, state, rewrite_function_name_map=rewrite_function_name_map):
        self.state = state
        self.function_name_map = rewrite_function_name_map
        self.typemap = state.type_annotation.typemap
        self.calltypes = state.type_annotation.calltypes

    def run(self):
        typingctx = self.state.typingctx

        # save array arg to call
        # call_varname -> array
        func_ir = self.state.func_ir
        blocks = func_ir.blocks
        saved_arr_arg = {}
        topo_order = find_topo_order(blocks)
        replaced = False

        for label in topo_order:
            block = blocks[label]
            new_body = []
            for stmt in block.body:
                if isinstance(stmt, ir.Assign) and isinstance(stmt.value, ir.Expr):
                    lhs = stmt.target.name
                    rhs = stmt.value
                    # replace A.func with np.func, and save A in saved_arr_arg
                    if (
                        rhs.op == "getattr"
                        and rhs.attr in self.function_name_map
                        and isinstance(
                            self.typemap[rhs.value.name], types.npytypes.Array
                        )
                    ):
                        rhs = stmt.value
                        arr = rhs.value
                        saved_arr_arg[lhs] = arr
                        scope = arr.scope
                        loc = arr.loc

                        g_dppy_var = ir.Var(scope, mk_unique_var("$load_global"), loc)
                        self.typemap[g_dppy_var.name] = types.misc.Module(numba_dppy)
                        g_dppy = ir.Global("numba_dppy", numba_dppy, loc)
                        g_dppy_assign = ir.Assign(g_dppy, g_dppy_var, loc)

                        dpnp_var = ir.Var(scope, mk_unique_var("$load_attr"), loc)
                        self.typemap[dpnp_var.name] = types.misc.Module(numba_dppy.dpnp)
                        getattr_dpnp = ir.Expr.getattr(g_dppy_var, "dpnp", loc)
                        dpnp_assign = ir.Assign(getattr_dpnp, dpnp_var, loc)

                        rhs.value = dpnp_var
                        new_body.append(g_dppy_assign)
                        new_body.append(dpnp_assign)

                        func_ir._definitions[g_dppy_var.name] = [getattr_dpnp]
                        func_ir._definitions[dpnp_var.name] = [getattr_dpnp]

                        # update func var type
                        func = getattr(numba_dppy.dpnp, rhs.attr)
                        func_typ = get_dpnp_func_typ(func)

                        self.typemap.pop(lhs)
                        self.typemap[lhs] = func_typ
                        replaced = True

                    if rhs.op == "call" and rhs.func.name in saved_arr_arg:
                        # add array as first arg
                        arr = saved_arr_arg[rhs.func.name]
                        # update call type signature to include array arg
                        old_sig = self.calltypes.pop(rhs)
                        # argsort requires kws for typing so sig.args can't be used
                        # reusing sig.args since some types become Const in sig
                        argtyps = old_sig.args[: len(rhs.args)]
                        kwtyps = {name: self.typemap[v.name] for name, v in rhs.kws}
                        self.calltypes[rhs] = self.typemap[rhs.func.name].get_call_type(
                            typingctx, [self.typemap[arr.name]] + list(argtyps), kwtyps
                        )
                        rhs.args = [arr] + rhs.args

                new_body.append(stmt)
            block.body = new_body
        return replaced


@register_pass(mutates_CFG=True, analysis_only=False)
class DPPYRewriteNdarrayFunctions(FunctionPass):
    _name = "dppy_rewrite_ndarray_functions_pass"

    def __init__(self):
        FunctionPass.__init__(self)

    def run_pass(self, state):
        rewrite_ndarray_function_name_pass = RewriteNdarrayFunctions(
            state, rewrite_function_name_map
        )

        mutated = rewrite_ndarray_function_name_pass.run()

        if mutated:
            remove_dead(state.func_ir.blocks, state.func_ir.arg_names, state.func_ir)
        state.func_ir.blocks = simplify_CFG(state.func_ir.blocks)

        return mutated



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dpctl
import numpy as np
import pytest

import numba_dppy as dppy
from numba_dppy.tests._helper import skip_test

list_of_filter_strs = [
    "opencl:gpu:0",
    "level_zero:gpu:0",
    "opencl:cpu:0",
]


@pytest.fixture(params=list_of_filter_strs)
def filter_str(request):
    return request.param


def test_caching_kernel_using_same_queue(filter_str):
    """Test kernel caching when the same queue is used to submit a kernel
    multiple times.

    Args:
        filter_str: SYCL filter selector string
    """
    if skip_test(filter_str):
        pytest.skip()
    global_size = 10
    N = global_size

    def data_parallel_sum(a, b, c):
        i = dppy.get_global_id(0)
        c[i] = a[i] + b[i]

    a = np.array(np.random.random(N), dtype=np.float32)
    b = np.array(np.random.random(N), dtype=np.float32)
    c = np.ones_like(a)

    with dpctl.device_context(filter_str) as gpu_queue:
        func = dppy.kernel(data_parallel_sum)
        cached_kernel = func[global_size, dppy.DEFAULT_LOCAL_SIZE].specialize(
            func._get_argtypes(a, b, c), gpu_queue
        )

        for i in range(10):
            _kernel = func[global_size, dppy.DEFAULT_LOCAL_SIZE].specialize(
                func._get_argtypes(a, b, c), gpu_queue
            )
            assert _kernel == cached_kernel


def test_caching_kernel_using_same_context(filter_str):
    """Test kernel caching for the scenario where different SYCL queues that
    share a SYCL context are used to submit a kernel.

    Args:
        filter_str: SYCL filter selector string
    """
    if skip_test(filter_str):
        pytest.skip()
    global_size = 10
    N = global_size

    def data_parallel_sum(a, b, c):
        i = dppy.get_global_id(0)
        c[i] = a[i] + b[i]

    a = np.array(np.random.random(N), dtype=np.float32)
    b = np.array(np.random.random(N), dtype=np.float32)
    c = np.ones_like(a)

    # Set the global queue to the default device so that the cached_kernel gets
    # created for that device
    dpctl.set_global_queue(filter_str)
    func = dppy.kernel(data_parallel_sum)
    default_queue = dpctl.get_current_queue()
    cached_kernel = func[global_size, dppy.DEFAULT_LOCAL_SIZE].specialize(
        func._get_argtypes(a, b, c), default_queue
    )
    for i in range(0, 10):
        # Each iteration create a fresh queue that will share the same context
        with dpctl.device_context(filter_str) as gpu_queue:
            _kernel = func[global_size, dppy.DEFAULT_LOCAL_SIZE].specialize(
                func._get_argtypes(a, b, c), gpu_queue
            )
            assert _kernel == cached_kernel



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module provides utilities to interact with USM memory."""

import dpctl
import dpctl.memory as dpctl_mem
import numpy as np

from numba_dppy import config

supported_numpy_dtype = [
    np.int32,
    np.int64,
    np.uint32,
    np.int64,
    np.float32,
    np.float64,
]


def has_usm_memory(obj):
    """
    Determine and return a SYCL device accesible object.

    as_usm_memory() converts Python object with `__sycl_usm_array_interface__`
    property to one of :class:`.MemoryUSMShared`, :class:`.MemoryUSMDevice`, or
    :class:`.MemoryUSMHost` instances. For more information please refer:
    https://github.com/IntelPython/dpctl/blob/0.8.0/dpctl/memory/_memory.pyx#L673

    Args:
        obj: Object to be tested and data copied from.

    Returns:
        A Python object allocated using USM memory if argument is already
        allocated using USM (zero-copy), None otherwise.
    """
    usm_mem = None
    try:
        usm_mem = dpctl_mem.as_usm_memory(obj)
    except Exception as e:
        if hasattr(obj, "base"):
            try:
                usm_mem = dpctl_mem.as_usm_memory(obj.base)
            except Exception as e:
                if config.DEBUG:
                    print(e)
        else:
            if config.DEBUG:
                print(e)

    return usm_mem


def copy_from_numpy_to_usm_obj(usm_allocated, obj):
    """
    Copy from supported objects to USM allocated data.

    This function copies the data of a supported Python type (only
    numpy.ndarray is supported at this point) into object that
    defines a __sycl_usm_array_interface__ attribute. For more information
    please refer to the specification of __sycl_usm_array_interface__:
    https://github.com/IntelPython/dpctl/wiki/Zero-copy-data-exchange-using-SYCL-USM#sycl-usm-array-interface

    Args:
        usm_allocated: An object that should define a
            __sycl_usm_array_interface__ dictionary. A TypeError is thrown
            if the object does not have such an attribute.
        obj (numpy.ndarray): Numpy ndarray, the data will be copied into.

    Raises:
        TypeError: If any argument is not of permitted type.
        ValueError:
            1. If size of data does not match.
            2. If obj is not C-contiguous.
    """
    usm_mem = has_usm_memory(usm_allocated)
    if usm_mem is None:
        raise TypeError("Source is not USM allocated.")

    if not isinstance(obj, np.ndarray):
        raise TypeError(
            "Obj is not USM allocated and is not of type "
            "numpy.ndarray. Obj type: %s" % (type(obj))
        )

    if obj.dtype not in [np.dtype(typ) for typ in supported_numpy_dtype]:
        raise ValueError(
            "dtype is not supprted. Supported dtypes "
            "are: %s" % (supported_numpy_dtype)
        )

    if not obj.flags.c_contiguous:
        raise ValueError("Only C-contiguous numpy.ndarray is currently supported!")

    size = np.prod(obj.shape)
    if usm_mem.size != (obj.dtype.itemsize * size):
        raise ValueError(
            "Size (Bytes) of data does not match. USM allocated "
            "memory size %d, supported object size: %d"
            % (usm_mem.size, (obj.dtype.itemsize * size))
        )

    obj_memview = memoryview(obj)
    obj_memview = obj_memview.cast("B")
    usm_mem.copy_from_host(obj_memview)


def copy_to_numpy_from_usm_obj(usm_allocated, obj):
    """
    Copy from USM allocated data to supported objects.

    Args:
        usm_allocated: An object that should define a
            __sycl_usm_array_interface__ dictionary. A TypeError is thrown
            if the object does not have such an attribute.
        obj (numpy.ndarray): Numpy ndarray, the data will be copied into.


    Raises:
        TypeError: If any argument is not of permitted type.
        ValueError: If size of data does not match.
    """
    usm_mem = has_usm_memory(usm_allocated)
    if usm_mem is None:
        raise TypeError("Source is not USM allocated.")

    if not isinstance(obj, np.ndarray):
        raise TypeError(
            "Obj is not USM allocated and is not of type "
            "numpy.ndarray. Obj type: %s" % (type(obj))
        )

    if obj.dtype not in [np.dtype(typ) for typ in supported_numpy_dtype]:
        raise ValueError(
            "dtype is not supprted. Supported dtypes "
            "are: %s" % (supported_numpy_dtype)
        )

    size = np.prod(obj.shape)
    if usm_mem.size != (obj.dtype.itemsize * size):
        raise ValueError(
            "Size (Bytes) of data does not match. USM allocated "
            "memory size %d, supported object size: %d"
            % (usm_mem.size, (obj.dtype.itemsize * size))
        )

    obj_memview = memoryview(obj)
    obj_memview = obj_memview.cast("B")
    usm_mem.copy_to_host(obj_memview)


def as_usm_obj(obj, queue=None, usm_type="shared", copy=True):
    """
    Determine and return a SYCL device accesible object.

    We try to determine if the provided object defines a valid
    __sycl_usm_array_interface__ dictionary.
    If not, we create a USM memory of `usm_type` and try to copy the data
    `obj` holds. Only numpy.ndarray is supported currently as `obj` if
    the object is not already allocated using USM.

    Args:
        obj: Object to be tested and data copied from.
        usm_type: USM type used in case obj is not already allocated using USM.
        queue (dpctl.SyclQueue): SYCL queue to be used to allocate USM
            memory in case obj is not already USM allocated.
        copy (bool): Flag to determine if we copy data from obj.

    Returns:
        A Python object allocated using USM memory.

    Raises:
        TypeError:
            1. If obj is not allocated on USM memory or is not of type
               numpy.ndarray, TypeError is raised.
            2. If queue is not of type dpctl.SyclQueue.
        ValueError:
            1. In case obj is not USM allocated, users need to pass
               the SYCL queue to be used for creating new memory. ValuieError
               is raised if queue argument is not provided.
            2. If usm_type is not valid.
            3. If dtype of the passed ndarray(obj) is not supported.
    """
    usm_mem = has_usm_memory(obj)

    if queue is None:
        raise ValueError(
            "Queue can not be None. Please provide the SYCL queue to be used."
        )
    if not isinstance(queue, dpctl.SyclQueue):
        raise TypeError(
            "queue has to be of dpctl.SyclQueue type. Got %s" % (type(queue))
        )

    if usm_mem is None:
        if not isinstance(obj, np.ndarray):
            raise TypeError(
                "Obj is not USM allocated and is not of type "
                "numpy.ndarray. Obj type: %s" % (type(obj))
            )

        if obj.dtype not in [np.dtype(typ) for typ in supported_numpy_dtype]:
            raise ValueError(
                "dtype is not supprted. Supported dtypes "
                "are: %s" % (supported_numpy_dtype)
            )

        size = np.prod(obj.shape)
        if usm_type == "shared":
            usm_mem = dpctl_mem.MemoryUSMShared(size * obj.dtype.itemsize, queue=queue)
        elif usm_type == "device":
            usm_mem = dpctl_mem.MemoryUSMDevice(size * obj.dtype.itemsize, queue=queue)
        elif usm_type == "host":
            usm_mem = dpctl_mem.MemoryUSMHost(size * obj.dtype.itemsize, queue=queue)
        else:
            raise ValueError(
                "Supported usm_type are: 'shared', "
                "'device' and 'host'. Provided: %s" % (usm_type)
            )

        if copy:
            # Copy data from numpy.ndarray
            copy_from_numpy_to_usm_obj(usm_mem, obj)

    return usm_mem



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numba import types
from numba.core.extending import overload, register_jitable
from numba.core.typing import signature

import numba_dppy.dpnp_glue as dpnp_lowering
import numba_dppy.dpnp_glue.dpnpimpl as dpnp_ext
from numba_dppy import dpctl_functions

from . import stubs


@register_jitable
def check_range(low, high):
    if low >= high:
        raise ValueError("Low cannot be >= High")


@register_jitable
def common_impl(low, high, res, dpnp_func, print_debug):
    check_range(low, high)

    sycl_queue = dpctl_functions.get_current_queue()
    res_usm = dpctl_functions.malloc_shared(res.size * res.itemsize, sycl_queue)

    dpnp_func(res_usm, low, high, res.size)

    event = dpctl_functions.queue_memcpy(
        sycl_queue, res.ctypes, res_usm, res.size * res.itemsize
    )
    dpctl_functions.event_wait(event)
    dpctl_functions.event_delete(event)
    dpctl_functions.free_with_queue(res_usm, sycl_queue)

    dpnp_ext._dummy_liveness_func([res.size])

    if print_debug:
        print("dpnp implementation")


@register_jitable
def common_impl_0_arg(res, dpnp_func, print_debug):
    sycl_queue = dpctl_functions.get_current_queue()
    res_usm = dpctl_functions.malloc_shared(res.size * res.itemsize, sycl_queue)

    dpnp_func(res_usm, res.size)

    event = dpctl_functions.queue_memcpy(
        sycl_queue, res.ctypes, res_usm, res.size * res.itemsize
    )
    dpctl_functions.event_wait(event)
    dpctl_functions.event_delete(event)
    dpctl_functions.free_with_queue(res_usm, sycl_queue)

    dpnp_ext._dummy_liveness_func([res.size])

    if print_debug:
        print("dpnp implementation")


@register_jitable
def common_impl_1_arg(arg1, res, dpnp_func, print_debug):
    sycl_queue = dpctl_functions.get_current_queue()
    res_usm = dpctl_functions.malloc_shared(res.size * res.itemsize, sycl_queue)

    try:
        dpnp_func(res_usm, arg1, res.size)
    except Exception:
        raise ValueError("Device not supported")

    event = dpctl_functions.queue_memcpy(
        sycl_queue, res.ctypes, res_usm, res.size * res.itemsize
    )
    dpctl_functions.event_wait(event)
    dpctl_functions.event_delete(event)
    dpctl_functions.free_with_queue(res_usm, sycl_queue)

    dpnp_ext._dummy_liveness_func([res.size])

    if print_debug:
        print("dpnp implementation")


@register_jitable
def common_impl_2_arg(arg1, arg2, res, dpnp_func, print_debug):
    sycl_queue = dpctl_functions.get_current_queue()
    res_usm = dpctl_functions.malloc_shared(res.size * res.itemsize, sycl_queue)

    dpnp_func(res_usm, arg1, arg2, res.size)

    event = dpctl_functions.queue_memcpy(
        sycl_queue, res.ctypes, res_usm, res.size * res.itemsize
    )
    dpctl_functions.event_wait(event)
    dpctl_functions.event_delete(event)
    dpctl_functions.free_with_queue(res_usm, sycl_queue)

    dpnp_ext._dummy_liveness_func([res.size])

    if print_debug:
        print("dpnp implementation")


@register_jitable
def common_impl_hypergeometric(ngood, nbad, nsample, res, dpnp_func, print_debug):
    sycl_queue = dpctl_functions.get_current_queue()
    res_usm = dpctl_functions.malloc_shared(res.size * res.itemsize, sycl_queue)

    dpnp_func(res_usm, ngood, nbad, nsample, res.size)

    event = dpctl_functions.queue_memcpy(
        sycl_queue, res.ctypes, res_usm, res.size * res.itemsize
    )
    dpctl_functions.event_wait(event)
    dpctl_functions.event_delete(event)

    dpctl_functions.free_with_queue(res_usm, sycl_queue)

    dpnp_ext._dummy_liveness_func([res.size])

    if print_debug:
        print("dpnp implementation")


@register_jitable
def common_impl_multinomial(n, pvals, res, dpnp_func, print_debug):
    sycl_queue = dpctl_functions.get_current_queue()
    res_usm = dpctl_functions.malloc_shared(res.size * res.itemsize, sycl_queue)

    pvals_usm = dpctl_functions.malloc_shared(pvals.size * pvals.itemsize, sycl_queue)
    event = dpctl_functions.queue_memcpy(
        sycl_queue, pvals_usm, pvals.ctypes, pvals.size * pvals.itemsize
    )
    dpctl_functions.event_wait(event)
    dpctl_functions.event_delete(event)

    dpnp_func(res_usm, n, pvals_usm, pvals.size, res.size)

    event = dpctl_functions.queue_memcpy(
        sycl_queue, res.ctypes, res_usm, res.size * res.itemsize
    )
    dpctl_functions.event_wait(event)
    dpctl_functions.event_delete(event)

    dpctl_functions.free_with_queue(res_usm, sycl_queue)
    dpctl_functions.free_with_queue(pvals_usm, sycl_queue)

    dpnp_ext._dummy_liveness_func([res.size])

    if print_debug:
        print("dpnp implementation")


@register_jitable
def common_impl_multivariate_normal(
    mean, cov, size, check_valid, tol, res, dpnp_func, print_debug
):
    sycl_queue = dpctl_functions.get_current_queue()
    res_usm = dpctl_functions.malloc_shared(res.size * res.itemsize, sycl_queue)

    mean_usm = dpctl_functions.malloc_shared(mean.size * mean.itemsize, sycl_queue)
    event = dpctl_functions.queue_memcpy(
        sycl_queue, mean_usm, mean.ctypes, mean.size * mean.itemsize
    )
    dpctl_functions.event_wait(event)
    dpctl_functions.event_delete(event)

    cov_usm = dpctl_functions.malloc_shared(cov.size * cov.itemsize, sycl_queue)
    event = dpctl_functions.queue_memcpy(
        sycl_queue, cov_usm, cov.ctypes, cov.size * cov.itemsize
    )
    dpctl_functions.event_wait(event)
    dpctl_functions.event_delete(event)

    dpnp_func(res_usm, mean.size, mean_usm, mean.size, cov_usm, cov.size, res.size)

    event = dpctl_functions.queue_memcpy(
        sycl_queue, res.ctypes, res_usm, res.size * res.itemsize
    )
    dpctl_functions.event_wait(event)
    dpctl_functions.event_delete(event)

    dpctl_functions.free_with_queue(res_usm, sycl_queue)
    dpctl_functions.free_with_queue(mean_usm, sycl_queue)
    dpctl_functions.free_with_queue(cov_usm, sycl_queue)

    dpnp_ext._dummy_liveness_func([res.size])

    if print_debug:
        print("dpnp implementation")


@overload(stubs.dpnp.random)
@overload(stubs.dpnp.sample)
@overload(stubs.dpnp.ranf)
@overload(stubs.dpnp.random_sample)
def dpnp_random_impl(size):
    name = "random_sample"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L391

    Function declaration:
    void custom_rng_uniform_c(void* result, long low, long high, size_t size)

    """
    sig = signature(ret_type, types.voidptr, types.int64, types.int64, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["float64", "NONE"], sig)

    res_dtype = np.float64

    PRINT_DEBUG = dpnp_lowering.DEBUG

    def dpnp_impl(size):
        res = np.empty(size, dtype=res_dtype)
        if res.size != 0:
            common_impl(0, 1, res, dpnp_func, PRINT_DEBUG)
        return res

    return dpnp_impl


@overload(stubs.dpnp.rand)
def dpnp_random_impl(*size):
    name = "random_sample"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L391

    Function declaration:
    void custom_rng_uniform_c(void* result, long low, long high, size_t size)

    """
    sig = signature(ret_type, types.voidptr, types.int64, types.int64, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["float64", "NONE"], sig)

    res_dtype = np.float64

    PRINT_DEBUG = dpnp_lowering.DEBUG

    def dpnp_impl(*size):
        res = np.empty(size, dtype=res_dtype)
        if res.size != 0:
            common_impl(0, 1, res, dpnp_func, PRINT_DEBUG)
        return res

    return dpnp_impl


@overload(stubs.dpnp.randint)
def dpnp_random_impl(low, high=None, size=None):
    name = "random_sample"
    dpnp_lowering.ensure_dpnp("randint")

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L391

    Function declaration:
    void custom_rng_uniform_c(void* result, long low, long high, size_t size)

    """
    sig = signature(ret_type, types.voidptr, types.int64, types.int64, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["int32", "NONE"], sig)

    res_dtype = np.int32

    PRINT_DEBUG = dpnp_lowering.DEBUG

    if size in (None, types.none):
        if high not in (None, types.none):

            def dpnp_impl(low, high=None, size=None):
                res = np.empty(1, dtype=res_dtype)
                common_impl(low, high, res, dpnp_func, PRINT_DEBUG)
                return res

        else:

            def dpnp_impl(low, high=None, size=None):
                res = np.empty(1, dtype=res_dtype)
                common_impl(0, low, res, dpnp_func, PRINT_DEBUG)
                return res

    else:
        if high not in (None, types.none):

            def dpnp_impl(low, high=None, size=None):
                res = np.empty(size, dtype=res_dtype)
                if res.size != 0:
                    common_impl(low, high, res, dpnp_func, PRINT_DEBUG)
                return res

        else:

            def dpnp_impl(low, high=None, size=None):
                res = np.empty(size, dtype=res_dtype)
                if res.size != 0:
                    common_impl(0, low, res, dpnp_func, PRINT_DEBUG)
                return res

    return dpnp_impl


@overload(stubs.dpnp.random_integers)
def dpnp_random_impl(low, high=None, size=None):
    name = "random_sample"
    dpnp_lowering.ensure_dpnp("random_integers")

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L391

    Function declaration:
    void custom_rng_uniform_c(void* result, long low, long high, size_t size)

    """
    sig = signature(ret_type, types.voidptr, types.int64, types.int64, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["int32", "NONE"], sig)

    res_dtype = np.int32

    PRINT_DEBUG = dpnp_lowering.DEBUG

    if size in (None, types.none):
        if high not in (None, types.none):

            def dpnp_impl(low, high=None, size=None):
                res = np.empty(1, dtype=res_dtype)
                common_impl(low, high + 1, res, dpnp_func, PRINT_DEBUG)
                return res

        else:

            def dpnp_impl(low, high=None, size=None):
                res = np.empty(1, dtype=res_dtype)
                common_impl(1, low + 1, res, dpnp_func, PRINT_DEBUG)
                return res

    else:
        if high not in (None, types.none):

            def dpnp_impl(low, high=None, size=None):
                res = np.empty(size, dtype=res_dtype)
                if res.size != 0:
                    common_impl(low, high + 1, res, dpnp_func, PRINT_DEBUG)
                return res

        else:

            def dpnp_impl(low, high=None, size=None):
                res = np.empty(size, dtype=res_dtype)
                if res.size != 0:
                    common_impl(1, low + 1, res, dpnp_func, PRINT_DEBUG)
                return res

    return dpnp_impl


@overload(stubs.dpnp.beta)
def dpnp_random_impl(a, b, size=None):
    name = "beta"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L36

    Function declaration:
    void custom_rng_beta_c(void* result, _DataType a, _DataType b, size_t size)

    """
    sig = signature(ret_type, types.voidptr, types.float64, types.float64, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["float64", "NONE"], sig)
    res_dtype = np.float64
    PRINT_DEBUG = dpnp_lowering.DEBUG

    if not (isinstance(a, types.Float)):
        raise ValueError("We only support float scalar for input: a")

    if not (isinstance(b, types.Float)):
        raise ValueError("We only support float scalar for input: b")

    if size in (None, types.none):

        def dpnp_impl(a, b, size=None):
            res = np.empty(1, dtype=res_dtype)
            common_impl_2_arg(a, b, res, dpnp_func, PRINT_DEBUG)
            return res[0]

    else:

        def dpnp_impl(a, b, size=None):
            res = np.empty(size, dtype=res_dtype)
            if res.size != 0:
                common_impl_2_arg(a, b, res, dpnp_func, PRINT_DEBUG)
            return res

    return dpnp_impl


@overload(stubs.dpnp.binomial)
def dpnp_random_impl(n, p, size=None):
    name = "binomial"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L56

    Function declaration:
    void custom_rng_binomial_c(void* result, int ntrial, double p, size_t size)

    """
    sig = signature(ret_type, types.voidptr, types.int32, types.float64, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["int32", "NONE"], sig)
    res_dtype = np.int32
    PRINT_DEBUG = dpnp_lowering.DEBUG

    if not (isinstance(n, types.Integer)):
        raise ValueError("We only support scalar for input: n")

    if not (isinstance(p, types.Float)):
        raise ValueError("We only support scalar for input: p")

    if size in (None, types.none):

        def dpnp_impl(n, p, size=None):
            res = np.empty(1, dtype=res_dtype)
            common_impl_2_arg(n, p, res, dpnp_func, PRINT_DEBUG)
            return res[0]

    else:

        def dpnp_impl(n, p, size=None):
            res = np.empty(size, dtype=res_dtype)
            if res.size != 0:
                common_impl_2_arg(n, p, res, dpnp_func, PRINT_DEBUG)
            return res

    return dpnp_impl


@overload(stubs.dpnp.chisquare)
def dpnp_random_impl(df, size=None):
    name = "chisquare"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L71

    Function declaration:
    void custom_rng_chi_square_c(void* result, int df, size_t size)
    """
    sig = signature(ret_type, types.voidptr, types.int32, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["float64", "NONE"], sig)
    res_dtype = np.float64
    PRINT_DEBUG = dpnp_lowering.DEBUG

    if not (isinstance(df, types.Integer)):
        raise ValueError("We only support scalar for input: df")

    if size in (None, types.none):

        def dpnp_impl(df, size=None):
            res = np.empty(1, dtype=res_dtype)
            common_impl_1_arg(df, res, dpnp_func, PRINT_DEBUG)
            return res[0]

    else:

        def dpnp_impl(df, size=None):
            res = np.empty(size, dtype=res_dtype)
            if res.size != 0:
                common_impl_1_arg(df, res, dpnp_func, PRINT_DEBUG)
            return res

    return dpnp_impl


@overload(stubs.dpnp.exponential)
def dpnp_random_impl(scale=1.0, size=None):
    name = "exponential"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L86

    Function declaration:
    void custom_rng_exponential_c(void* result, _DataType beta, size_t size)
    """
    sig = signature(ret_type, types.voidptr, types.float64, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["float64", "NONE"], sig)
    res_dtype = np.float64
    PRINT_DEBUG = dpnp_lowering.DEBUG

    if not isinstance(scale, float):
        if not (isinstance(scale, types.Float)):
            raise ValueError("We only support scalar for input: scale")

    if size in (None, types.none):

        def dpnp_impl(scale=1.0, size=None):
            res = np.empty(1, dtype=res_dtype)
            common_impl_1_arg(scale, res, dpnp_func, PRINT_DEBUG)
            return res[0]

    else:

        def dpnp_impl(scale=1.0, size=None):
            res = np.empty(size, dtype=res_dtype)
            if res.size != 0:
                common_impl_1_arg(scale, res, dpnp_func, PRINT_DEBUG)
            return res

    return dpnp_impl


@overload(stubs.dpnp.gamma)
def dpnp_random_impl(shape, scale=1.0, size=None):
    name = "gamma"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L105

    Function declaration:
    void custom_rng_gamma_c(void* result, _DataType shape, _DataType scale, size_t size)
    """
    sig = signature(ret_type, types.voidptr, types.float64, types.float64, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["float64", "NONE"], sig)
    res_dtype = np.float64
    PRINT_DEBUG = dpnp_lowering.DEBUG

    if not (isinstance(shape, types.Float)):
        raise ValueError("We only support scalar for input: shape")

    if not isinstance(scale, float):
        if not (isinstance(scale, types.Float)):
            raise ValueError("We only support scalar for input: scale")

    if size in (None, types.none):

        def dpnp_impl(shape, scale=1.0, size=None):
            res = np.empty(1, dtype=res_dtype)
            common_impl_2_arg(shape, scale, res, dpnp_func, PRINT_DEBUG)
            return res[0]

    else:

        def dpnp_impl(shape, scale=1.0, size=None):
            res = np.empty(size, dtype=res_dtype)
            if res.size != 0:
                common_impl_2_arg(shape, scale, res, dpnp_func, PRINT_DEBUG)

    return dpnp_impl


@overload(stubs.dpnp.geometric)
def dpnp_random_impl(p, size=None):
    name = "geometric"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L139

    Function declaration:
    void custom_rng_geometric_c(void* result, float p, size_t size)
    """
    sig = signature(ret_type, types.voidptr, types.float32, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["int32", "NONE"], sig)
    res_dtype = np.int32
    PRINT_DEBUG = dpnp_lowering.DEBUG

    if not (isinstance(p, types.Float)):
        raise ValueError("We only support scalar for input: p")

    if size in (None, types.none):

        def dpnp_impl(p, size=None):
            res = np.empty(1, dtype=res_dtype)
            common_impl_1_arg(p, res, dpnp_func, PRINT_DEBUG)
            return res[0]

    else:

        def dpnp_impl(p, size=None):
            res = np.empty(size, dtype=res_dtype)
            if res.size != 0:
                common_impl_1_arg(p, res, dpnp_func, PRINT_DEBUG)
            return res

    return dpnp_impl


@overload(stubs.dpnp.gumbel)
def dpnp_random_impl(loc=0.0, scale=1.0, size=None):
    name = "gumbel"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L154

    Function declaration:
    void custom_rng_gumbel_c(void* result, double loc, double scale, size_t size)
    """
    sig = signature(ret_type, types.voidptr, types.float64, types.float64, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["float64", "NONE"], sig)
    res_dtype = np.float64
    PRINT_DEBUG = dpnp_lowering.DEBUG

    if not isinstance(loc, float):
        if not (isinstance(loc, types.Float)):
            raise ValueError("We only support scalar for input: loc")

    if not isinstance(scale, float):
        if not (isinstance(scale, types.Float)):
            raise ValueError("We only support scalar for input: scale")

    if size in (None, types.none):

        def dpnp_impl(loc=0.0, scale=1.0, size=None):
            res = np.empty(1, dtype=res_dtype)
            common_impl_2_arg(loc, scale, res, dpnp_func, PRINT_DEBUG)
            return res[0]

    else:

        def dpnp_impl(loc=0.0, scale=1.0, size=None):
            res = np.empty(size, dtype=res_dtype)
            if res.size != 0:
                common_impl_2_arg(loc, scale, res, dpnp_func, PRINT_DEBUG)
            return res

    return dpnp_impl


@overload(stubs.dpnp.hypergeometric)
def dpnp_random_impl(ngood, nbad, nsample, size=None):
    name = "hypergeometric"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L169

    Function declaration:
    void custom_rng_hypergeometric_c(void* result, int l, int s, int m, size_t size)
    """
    sig = signature(
        ret_type, types.voidptr, types.int32, types.int32, types.int32, types.intp
    )
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["int32", "NONE"], sig)

    res_dtype = np.int32

    PRINT_DEBUG = dpnp_lowering.DEBUG

    if not (isinstance(ngood, types.Integer)):
        raise ValueError("We only support scalar for input: ngood")

    if not (isinstance(nbad, types.Integer)):
        raise ValueError("We only support scalar for input: nbad")

    if not (isinstance(nsample, types.Integer)):
        raise ValueError("We only support scalar for input: nsample")

    if size in (None, types.none):

        def dpnp_impl(ngood, nbad, nsample, size=None):
            res = np.empty(1, dtype=res_dtype)
            common_impl_hypergeometric(
                ngood, nbad, nsample, res, dpnp_func, PRINT_DEBUG
            )
            return res[0]

    else:

        def dpnp_impl(ngood, nbad, nsample, size=None):
            res = np.empty(size, dtype=res_dtype)
            if res.size != 0:
                common_impl_hypergeometric(
                    ngood, nbad, nsample, res, dpnp_func, PRINT_DEBUG
                )
            return res

    return dpnp_impl


@overload(stubs.dpnp.laplace)
def dpnp_random_impl(loc=0.0, scale=1.0, size=None):
    name = "laplace"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L184

    Function declaration:
    void custom_rng_laplace_c(void* result, double loc, double scale, size_t size)
    """
    sig = signature(ret_type, types.voidptr, types.float64, types.float64, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["float64", "NONE"], sig)
    res_dtype = np.float64
    PRINT_DEBUG = dpnp_lowering.DEBUG

    if not isinstance(loc, float):
        if not (isinstance(loc, types.Float)):
            raise ValueError("We only support scalar for input: loc")

    if not isinstance(scale, float):
        if not (isinstance(scale, types.Float)):
            raise ValueError("We only support scalar for input: scale")

    if size in (None, types.none):

        def dpnp_impl(loc=0.0, scale=1.0, size=None):
            res = np.empty(1, dtype=res_dtype)
            common_impl_2_arg(loc, scale, res, dpnp_func, PRINT_DEBUG)
            return res[0]

    else:

        def dpnp_impl(loc=0.0, scale=1.0, size=None):
            res = np.empty(size, dtype=res_dtype)
            if res.size != 0:
                common_impl_2_arg(loc, scale, res, dpnp_func, PRINT_DEBUG)
            return res

    return dpnp_impl


@overload(stubs.dpnp.lognormal)
def dpnp_random_impl(mean=0.0, sigma=1.0, size=None):
    name = "lognormal"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L199

    Function declaration:
    void custom_rng_lognormal_c(void* result, _DataType mean, _DataType stddev, size_t size)
    """
    sig = signature(ret_type, types.voidptr, types.float64, types.float64, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["float64", "NONE"], sig)
    res_dtype = np.float64
    PRINT_DEBUG = dpnp_lowering.DEBUG

    if not isinstance(mean, float):
        if not (isinstance(mean, types.Float)):
            raise ValueError("We only support scalar for input: loc")

    if not isinstance(sigma, float):
        if not (isinstance(sigma, types.Float)):
            raise ValueError("We only support scalar for input: scale")

    if size in (None, types.none):

        def dpnp_impl(mean=0.0, sigma=1.0, size=None):
            res = np.empty(1, dtype=res_dtype)
            common_impl_2_arg(mean, sigma, res, dpnp_func, PRINT_DEBUG)
            return res[0]

    else:

        def dpnp_impl(mean=0.0, sigma=1.0, size=None):
            res = np.empty(size, dtype=res_dtype)
            if res.size != 0:
                common_impl_2_arg(mean, sigma, res, dpnp_func, PRINT_DEBUG)
            return res

    return dpnp_impl


@overload(stubs.dpnp.multinomial)
def dpnp_random_impl(n, pvals, size=None):
    name = "multinomial"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L218

    Function declaration:
    void custom_rng_multinomial_c(void* result, int ntrial, const double* p_vector,
                                  const size_t p_vector_size, size_t size)
    """
    sig = signature(
        ret_type, types.voidptr, types.int32, types.voidptr, types.intp, types.intp
    )
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["int32", "NONE"], sig)

    res_dtype = np.int32

    PRINT_DEBUG = dpnp_lowering.DEBUG

    if not isinstance(n, types.Integer):
        raise TypeError(
            "np.random.multinomial(): n should be an " "integer, got %s" % (n,)
        )

    if not isinstance(pvals, (types.Sequence, types.Array)):
        raise TypeError(
            "np.random.multinomial(): pvals should be an "
            "array or sequence, got %s" % (pvals,)
        )

    if size in (None, types.none):

        def dpnp_impl(n, pvals, size=None):
            out = np.zeros(len(pvals), res_dtype)
            common_impl_multinomial(n, pvals, out, dpnp_func, PRINT_DEBUG)
            return out

    elif isinstance(size, types.Integer):

        def dpnp_impl(n, pvals, size=None):
            out = np.zeros((size, len(pvals)), res_dtype)
            common_impl_multinomial(n, pvals, out, dpnp_func, PRINT_DEBUG)
            return out

    elif isinstance(size, types.BaseTuple):

        def dpnp_impl(n, pvals, size=None):
            out = np.zeros(size + (len(pvals),), res_dtype)
            common_impl_multinomial(n, pvals, out, dpnp_func, PRINT_DEBUG)
            return out

    else:
        raise TypeError(
            "np.random.multinomial(): size should be int or "
            "tuple or None, got %s" % (size,)
        )

    return dpnp_impl


@overload(stubs.dpnp.multivariate_normal)
def dpnp_random_impl(mean, cov, size=None, check_valid="warn", tol=1e-8):
    name = "multivariate_normal"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L239

    Function declaration:
    void custom_rng_multivariate_normal_c(void* result,
                                      const int dimen,
                                      const double* mean_vector,
                                      const size_t mean_vector_size,
                                      const double* cov_vector,
                                      const size_t cov_vector_size,
                                      size_t size)
    """
    sig = signature(
        ret_type,
        types.voidptr,
        types.int32,
        types.voidptr,
        types.intp,
        types.voidptr,
        types.intp,
        types.intp,
    )
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["float64", "NONE"], sig)

    res_dtype = np.float64

    PRINT_DEBUG = dpnp_lowering.DEBUG

    if size in (None, types.none):

        def dpnp_impl(mean, cov, size=None, check_valid="warn", tol=1e-8):
            out = np.empty(mean.shape, dtype=res_dtype)
            common_impl_multivariate_normal(
                mean, cov, size, check_valid, tol, out, dpnp_func, PRINT_DEBUG
            )
            return out

    elif isinstance(size, types.Integer):

        def dpnp_impl(mean, cov, size=None, check_valid="warn", tol=1e-8):
            new_size = (size,)
            new_size = new_size + (mean.size,)
            out = np.empty(new_size, dtype=res_dtype)
            common_impl_multivariate_normal(
                mean, cov, size, check_valid, tol, out, dpnp_func, PRINT_DEBUG
            )
            return out

    elif isinstance(size, types.BaseTuple):

        def dpnp_impl(mean, cov, size=None, check_valid="warn", tol=1e-8):
            new_size = size + (mean.size,)
            out = np.empty(new_size, dtype=res_dtype)
            common_impl_multivariate_normal(
                mean, cov, size, check_valid, tol, out, dpnp_func, PRINT_DEBUG
            )
            return out

    else:
        raise TypeError(
            "np.random.multivariate_normal(): size should be int or "
            "tuple or None, got %s" % (size,)
        )

    return dpnp_impl


@overload(stubs.dpnp.negative_binomial)
def dpnp_random_impl(n, p, size=None):
    name = "negative_binomial"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L267

    Function declaration:
    void custom_rng_negative_binomial_c(void* result, double a, double p, size_t size)
    """
    sig = signature(ret_type, types.voidptr, types.int32, types.float64, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["int32", "NONE"], sig)
    res_dtype = np.int32
    PRINT_DEBUG = dpnp_lowering.DEBUG

    if not (isinstance(n, types.Integer)):
        raise ValueError("We only support scalar for input: n")

    if not (isinstance(p, types.Float)):
        raise ValueError("We only support scalar for input: p")

    if size in (None, types.none):

        def dpnp_impl(n, p, size=None):
            res = np.empty(1, dtype=res_dtype)
            common_impl_2_arg(n, p, res, dpnp_func, PRINT_DEBUG)
            return res[0]

    else:

        def dpnp_impl(n, p, size=None):
            res = np.empty(size, dtype=res_dtype)
            if res.size != 0:
                common_impl_2_arg(n, p, res, dpnp_func, PRINT_DEBUG)
            return res

    return dpnp_impl


@overload(stubs.dpnp.normal)
def dpnp_random_impl(loc=0.0, scale=1.0, size=None):
    name = "normal"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L282

    Function declaration:
    void custom_rng_normal_c(void* result, _DataType mean, _DataType stddev, size_t size)
    """
    sig = signature(ret_type, types.voidptr, types.float64, types.float64, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["float64", "NONE"], sig)
    res_dtype = np.float64
    PRINT_DEBUG = dpnp_lowering.DEBUG

    if not isinstance(loc, float):
        if not (isinstance(loc, types.Float)):
            raise ValueError("We only support scalar for input: loc")

    if not isinstance(scale, float):
        if not (isinstance(scale, types.Float)):
            raise ValueError("We only support scalar for input: scale")

    if size in (None, types.none):

        def dpnp_impl(loc=0.0, scale=1.0, size=None):
            res = np.empty(1, dtype=res_dtype)
            common_impl_2_arg(loc, scale, res, dpnp_func, PRINT_DEBUG)
            return res[0]

    else:

        def dpnp_impl(loc=0.0, scale=1.0, size=None):
            res = np.empty(size, dtype=res_dtype)
            if res.size != 0:
                common_impl_2_arg(loc, scale, res, dpnp_func, PRINT_DEBUG)
            return res

    return dpnp_impl


@overload(stubs.dpnp.poisson)
def dpnp_random_impl(lam=1.0, size=None):
    name = "poisson"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L297

    Function declaration:
    void custom_rng_poisson_c(void* result, double lambda, size_t size)
    """
    sig = signature(ret_type, types.voidptr, types.float64, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["int32", "NONE"], sig)
    res_dtype = np.int32
    PRINT_DEBUG = dpnp_lowering.DEBUG

    if not isinstance(lam, float):
        if not (isinstance(lam, types.Float)):
            raise ValueError("We only support scalar for input: lam")

    if size in (None, types.none):

        def dpnp_impl(lam=1.0, size=None):
            res = np.empty(1, dtype=res_dtype)
            common_impl_1_arg(lam, res, dpnp_func, PRINT_DEBUG)
            return res[0]

    else:

        def dpnp_impl(lam=1.0, size=None):
            res = np.empty(size, dtype=res_dtype)
            if res.size != 0:
                common_impl_1_arg(lam, res, dpnp_func, PRINT_DEBUG)
            return res

    return dpnp_impl


@overload(stubs.dpnp.rayleigh)
def dpnp_random_impl(scale=1.0, size=None):
    name = "rayleigh"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L312

    Function declaration:
    void custom_rng_rayleigh_c(void* result, _DataType scale, size_t size)
    """
    sig = signature(ret_type, types.voidptr, types.float64, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["float64", "NONE"], sig)
    res_dtype = np.float64
    PRINT_DEBUG = dpnp_lowering.DEBUG

    if not isinstance(scale, float):
        if not (isinstance(scale, types.Float)):
            raise ValueError("We only support scalar for input: scale")

    if size in (None, types.none):

        def dpnp_impl(scale=1.0, size=None):
            res = np.empty(1, dtype=res_dtype)
            common_impl_1_arg(scale, res, dpnp_func, PRINT_DEBUG)
            return res[0]

    else:

        def dpnp_impl(scale=1.0, size=None):
            res = np.empty(size, dtype=res_dtype)
            if res.size != 0:
                common_impl_1_arg(scale, res, dpnp_func, PRINT_DEBUG)
            return res

    return dpnp_impl


@overload(stubs.dpnp.standard_cauchy)
def dpnp_random_impl(size=None):
    name = "standard_cauchy"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L331

    Function declaration:
    void custom_rng_standard_cauchy_c(void* result, size_t size)
    """
    sig = signature(ret_type, types.voidptr, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["float64", "NONE"], sig)
    res_dtype = np.float64
    PRINT_DEBUG = dpnp_lowering.DEBUG

    if size in (None, types.none):

        def dpnp_impl(size=None):
            res = np.empty(1, dtype=res_dtype)
            common_impl_0_arg(res, dpnp_func, PRINT_DEBUG)
            return res[0]

    else:

        def dpnp_impl(size=None):
            res = np.empty(size, dtype=res_dtype)
            if res.size != 0:
                common_impl_0_arg(res, dpnp_func, PRINT_DEBUG)
            return res

    return dpnp_impl


@overload(stubs.dpnp.standard_exponential)
def dpnp_random_impl(size=None):
    name = "standard_exponential"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L350

    Function declaration:
    void custom_rng_standard_exponential_c(void* result, size_t size)
    """
    sig = signature(ret_type, types.voidptr, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["float64", "NONE"], sig)
    res_dtype = np.float64
    PRINT_DEBUG = dpnp_lowering.DEBUG

    if size in (None, types.none):

        def dpnp_impl(size=None):
            res = np.empty(1, dtype=res_dtype)
            common_impl_0_arg(res, dpnp_func, PRINT_DEBUG)
            return res[0]

    else:

        def dpnp_impl(size=None):
            res = np.empty(size, dtype=res_dtype)
            if res.size != 0:
                common_impl_0_arg(res, dpnp_func, PRINT_DEBUG)
            return res

    return dpnp_impl


@overload(stubs.dpnp.standard_gamma)
def dpnp_random_impl(shape, size=None):
    name = "standard_gamma"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L364

    Function declaration:
    void custom_rng_standard_gamma_c(void* result, _DataType shape, size_t size)
    """
    sig = signature(ret_type, types.voidptr, types.float64, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["float64", "NONE"], sig)
    res_dtype = np.float64
    PRINT_DEBUG = dpnp_lowering.DEBUG

    if not (isinstance(shape, types.Float)):
        raise ValueError("We only support scalar for input: shape")

    if size in (None, types.none):

        def dpnp_impl(shape, size=None):
            res = np.empty(1, dtype=res_dtype)
            common_impl_1_arg(shape, res, dpnp_func, PRINT_DEBUG)
            return res[0]

    else:

        def dpnp_impl(shape, size=None):
            res = np.empty(size, dtype=res_dtype)
            if res.size != 0:
                common_impl_1_arg(shape, res, dpnp_func, PRINT_DEBUG)
            return res

    return dpnp_impl


@overload(stubs.dpnp.standard_normal)
def dpnp_random_impl(size=None):
    name = "normal"
    dpnp_lowering.ensure_dpnp("standard_normal")

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L282

    Function declaration:
    void custom_rng_normal_c(void* result, _DataType mean, _DataType stddev, size_t size)
    """
    sig = signature(ret_type, types.voidptr, types.float64, types.float64, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["float64", "NONE"], sig)
    res_dtype = np.float64
    PRINT_DEBUG = dpnp_lowering.DEBUG

    if size in (None, types.none):

        def dpnp_impl(size=None):
            res = np.empty(1, dtype=res_dtype)
            common_impl(0.0, 1.0, res, dpnp_func, PRINT_DEBUG)
            return res[0]

    else:

        def dpnp_impl(size=None):
            res = np.empty(size, dtype=res_dtype)
            if res.size != 0:
                common_impl(0.0, 1.0, res, dpnp_func, PRINT_DEBUG)
            return res

    return dpnp_impl


@overload(stubs.dpnp.uniform)
def dpnp_random_impl(low=0.0, high=1.0, size=None):
    name = "uniform"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L391

    Function declaration:
    void custom_rng_uniform_c(void* result, long low, long high, size_t size)
    """
    sig = signature(ret_type, types.voidptr, types.int64, types.int64, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["float64", "NONE"], sig)

    res_dtype = np.float64

    PRINT_DEBUG = dpnp_lowering.DEBUG

    if size in (None, types.none):

        def dpnp_impl(low=0.0, high=1.0, size=None):
            res = np.empty(1, dtype=res_dtype)
            common_impl(low, high, res, dpnp_func, PRINT_DEBUG)
            return res

    else:

        def dpnp_impl(low=0.0, high=1.0, size=None):
            res = np.empty(size, dtype=res_dtype)
            if res.size != 0:
                common_impl(low, high, res, dpnp_func, PRINT_DEBUG)
            return res

    return dpnp_impl


@overload(stubs.dpnp.weibull)
def dpnp_random_impl(a, size=None):
    name = "weibull"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.4.0/dpnp/backend/custom_kernels_random.cpp#L411

    Function declaration:
    void custom_rng_weibull_c(void* result, double alpha, size_t size)
    """
    sig = signature(ret_type, types.voidptr, types.float64, types.intp)
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, ["float64", "NONE"], sig)
    res_dtype = np.float64
    PRINT_DEBUG = dpnp_lowering.DEBUG

    if not (isinstance(a, types.Float)):
        raise ValueError("We only support scalar for input: a")

    if size in (None, types.none):

        def dpnp_impl(a, size=None):
            res = np.empty(1, dtype=res_dtype)
            common_impl_1_arg(a, res, dpnp_func, PRINT_DEBUG)
            return res[0]

    else:

        def dpnp_impl(a, size=None):
            res = np.empty(size, dtype=res_dtype)
            if res.size != 0:
                common_impl_1_arg(a, res, dpnp_func, PRINT_DEBUG)
            return res

    return dpnp_impl



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Re export
from .ocl.stubs import (
    CLK_GLOBAL_MEM_FENCE,
    CLK_LOCAL_MEM_FENCE,
    atomic,
    barrier,
    get_global_id,
    get_global_size,
    get_group_id,
    get_local_id,
    get_local_size,
    get_num_groups,
    get_work_dim,
    local,
    mem_fence,
    sub_group_barrier,
)

"""
We are importing dpnp stub module to make Numba recognize the
module when we rename Numpy functions.
"""
from .dpnp_glue.stubs import dpnp

DEFAULT_LOCAL_SIZE = []

import dpctl

from . import initialize, target
from .decorators import autojit, func, kernel


def is_available():
    """Returns a boolean indicating if dpctl could find a default device.

    A valueError is thrown by dpctl if no default device is found and it
    implies that numba-dppy cannot create a SYCL queue to compile kernels.

    Returns:
        bool: True if a default SYCL device is found, otherwise False.
    """
    try:
        d = dpctl.select_default_device()
        return not d.is_host
    except ValueError:
        return False


initialize.load_dpctl_sycl_interface()



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from llvmlite import ir
from numba.core.debuginfo import DIBuilder


class DPPYDIBuilder(DIBuilder):
    def __init__(self, module, filepath, linkage_name):
        DIBuilder.__init__(self, module, filepath)
        self.linkage_name = linkage_name

    def mark_subprogram(self, function, name, line):
        di_subp = self._add_subprogram(
            name=name, linkagename=self.linkage_name, line=line
        )
        function.set_metadata("dbg", di_subp)
        # disable inlining for this function for easier debugging
        function.attributes.add("noinline")

    def _di_compile_unit(self):
        di = super()._di_compile_unit()
        operands = dict(di.operands)
        operands["language"] = ir.DIToken("DW_LANG_C_plus_plus")
        operands["producer"] = "numba-dppy"
        di.operands = tuple(operands.items())
        return di



#! /usr/bin/env python
# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pathlib
import shutil
import sys

import pytest

import numba_dppy
from numba_dppy import config

pexpect = pytest.importorskip("pexpect")

pytestmark = pytest.mark.skipif(
    not shutil.which("gdb-oneapi"),
    reason="Intel® Distribution for GDB* is not available",
)


# TODO: go to helper
class gdb:
    def __init__(self):
        self.spawn()
        self.setup_gdb()

    def __del__(self):
        self.teardown_gdb()

    def spawn(self):
        env = os.environ.copy()
        env["NUMBA_OPT"] = "0"

        self.child = pexpect.spawn("gdb-oneapi -q python", env=env, encoding="utf-8")
        if config.DEBUG:
            self.child.logfile = sys.stdout

    def setup_gdb(self):
        self.child.expect("(gdb)", timeout=5)
        self.child.sendline("set breakpoint pending on")
        self.child.expect("(gdb)", timeout=5)
        self.child.sendline("set style enabled off")  # disable colors symbols

    def teardown_gdb(self):
        self.child.expect("(gdb)", timeout=5)
        self.child.sendline("quit")
        self.child.expect("Quit anyway?", timeout=5)
        self.child.sendline("y")

    def _command(self, command):
        self.child.expect("(gdb)", timeout=5)
        self.child.sendline(command)

    def breakpoint(self, breakpoint):
        self._command("break " + breakpoint)

    def run(self, script):
        self._command("run " + self.script_path(script))

    def backtrace(self):
        self._command("backtrace")

    def print(self, var):
        self._command("print " + var)

    def info_functions(self, function):
        self._command("info functions " + function)

    def info_locals(self):
        self._command("info locals")

    def next(self):
        self._command("next")

    def ptype(self, var):
        self._command("ptype " + var)

    def whatis(self, var):
        self._command("whatis " + var)

    def step(self):
        self._command("step")

    def stepi(self):
        self._command("stepi")

    @staticmethod
    def script_path(script):
        package_path = pathlib.Path(numba_dppy.__file__).parent
        return str(package_path / "examples/debug" / script)


@pytest.fixture
def app():
    return gdb()


@pytest.mark.parametrize("api", ["numba", "numba-dppy"])
def test_breakpoint_row_number(app, api):
    app.breakpoint("dppy_numba_basic.py:25")
    app.run("dppy_numba_basic.py --api={api}".format(api=api))

    app.child.expect(r"Thread .* hit Breakpoint .* at dppy_numba_basic.py:25")
    app.child.expect(r"25\s+param_c = param_a \+ 10")


# commands/backtrace
def test_backtrace(app):
    app.breakpoint("simple_dppy_func.py:23")
    app.run("simple_dppy_func.py")

    app.child.expect(r"Thread .* hit Breakpoint .* at simple_dppy_func.py:23")
    app.child.expect(r"23\s+result = a_in_func \+ b_in_func")

    app.backtrace()

    app.child.expect(r"#0.*__main__::func_sum .* at simple_dppy_func.py:23")
    app.child.expect(r"#1.*__main__::kernel_sum .* at simple_dppy_func.py:30")


# commands/break_conditional
def test_break_conditional(app):
    app.breakpoint("simple_sum.py:24 if i == 1")
    app.run("simple_sum.py")

    app.child.expect(r"Thread .* hit Breakpoint .* at simple_sum.py:24")
    app.child.expect(r"24\s+c\[i\] = a\[i\] \+ b\[i\]")

    app.print("i")

    app.child.expect(r"\$1 = 1")


# commands/break_file_func
def test_break_file_function(app):
    app.breakpoint("simple_sum.py:data_parallel_sum")
    app.run("simple_sum.py")

    app.child.expect(r"Thread .* hit Breakpoint .* at simple_sum.py:23")
    app.child.expect(r"23\s+i = dppy.get_global_id\(0\)")


# commands/break_func
def test_break_function(app):
    app.breakpoint("data_parallel_sum")
    app.run("simple_sum.py")

    app.child.expect(r"Thread .* hit Breakpoint .* at simple_sum.py:23")
    app.child.expect(r"23\s+i = dppy.get_global_id\(0\)")


# commands/break_nested_func
def test_break_nested_function(app):
    app.breakpoint("simple_dppy_func.py:func_sum")
    app.run("simple_dppy_func.py")

    app.child.expect(r"Thread .* hit Breakpoint .* at simple_dppy_func.py:23")
    app.child.expect(r"23\s+result = a_in_func \+ b_in_func")


# commands/info_func
def test_info_functions(app):
    app.breakpoint("simple_sum.py:22")
    app.run("simple_sum.py")

    app.child.expect(r"Thread .* hit Breakpoint .* at simple_sum.py:23")
    app.child.expect(r"23\s+i = dppy.get_global_id\(0\)")

    app.info_functions("data_parallel_sum")

    app.child.expect(r"21:\s+void __main__::data_parallel_sum\(.*\);")


# commands/local_variables_0
def test_local_variables(app):
    app.breakpoint("sum_local_vars.py:26")
    app.run("sum_local_vars.py")

    app.child.expect(r"Thread .* hit Breakpoint .* at sum_local_vars.py:26")
    app.child.expect(r"26\s+c\[i\] = l1 \+ l2")

    app.info_locals()

    app.child.expect(r"a =")
    app.child.expect(r"b =")
    app.child.expect(r"c =")
    app.child.expect(r"i = 0")
    app.child.expect(r"l1 = [0-9]\.[0-9]{3}")
    app.child.expect(r"l2 = [0-9]\.[0-9]{3}")

    app.print("a")
    app.child.expect(r"\$1 = '\\000' \<repeats 55 times\>")

    app.print("l1")
    app.child.expect(r"\$2 = [0-9]\.[0-9]{3}")

    app.print("l2")
    app.child.expect(r"\$3 = [0-9]\.[0-9]{3}")

    app.ptype("a")
    app.child.expect(r"type = byte \[56\]")

    app.whatis("a")
    app.child.expect(r"type = byte \[56\]")

    app.ptype("l1")
    app.child.expect(r"type = double")

    app.whatis("l1")
    app.child.expect(r"type = double")


# commands/next
def test_next(app):
    app.breakpoint("simple_dppy_func.py:30")
    app.run("simple_dppy_func.py")

    app.child.expect(r"Thread .* hit Breakpoint .* at simple_dppy_func.py:30")
    app.child.expect(
        r"30\s+c_in_kernel\[i\] = func_sum\(a_in_kernel\[i\], b_in_kernel\[i\]\)"
    )

    app.next()
    app.next()

    app.child.expect(r"Done\.\.\.")


# commands/step_dppy_func
def test_step(app):
    app.breakpoint("simple_dppy_func.py:30")
    app.run("simple_dppy_func.py")
    app.child.expect(r"Thread .* hit Breakpoint .* at simple_dppy_func.py:30")
    app.child.expect(
        r"30\s+c_in_kernel\[i\] = func_sum\(a_in_kernel\[i\], b_in_kernel\[i\]\)"
    )

    app.step()
    app.step()

    app.child.expect(r"__main__::func_sum \(\) at simple_dppy_func.py:23")
    app.child.expect(r"23\s+result = a_in_func \+ b_in_func")


# commands/stepi
def test_stepi(app):
    app.breakpoint("simple_dppy_func.py:30")
    app.run("simple_dppy_func.py")

    app.child.expect(r"Thread .* hit Breakpoint .* at simple_dppy_func.py:30")
    app.child.expect(
        r"30\s+c_in_kernel\[i\] = func_sum\(a_in_kernel\[i\], b_in_kernel\[i\]\)"
    )

    app.stepi()

    app.child.expect(
        r"0x[0-f]+\s+30\s+c_in_kernel\[i\] = func_sum\(a_in_kernel\[i\], b_in_kernel\[i\]\)"
    )

    app.stepi()

    app.child.expect(r"Switching to Thread")
    app.child.expect(r"Thread .* hit Breakpoint .* at simple_dppy_func.py:30")
    app.child.expect(
        r"30\s+c_in_kernel\[i\] = func_sum\(a_in_kernel\[i\], b_in_kernel\[i\]\)"
    )



################################################################################
#                                 Numba-DPPY
#
# Copyright 2020-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import dpctl
import numpy as np
import pytest
from numba import njit

import numba_dppy as dppy
from numba_dppy.tests._helper import assert_auto_offloading, is_gen12, skip_test

list_of_filter_strs = [
    "opencl:gpu:0",
    "level_zero:gpu:0",
    "opencl:cpu:0",
]


@pytest.fixture(params=list_of_filter_strs)
def filter_str(request):
    return request.param


list_of_binary_ops = [
    "add",
    "subtract",
    "multiply",
    "divide",
    "true_divide",
    "power",
    "remainder",
    "mod",
    "fmod",
    "hypot",
    "maximum",
    "minimum",
    "fmax",
    "fmin",
]


@pytest.fixture(params=list_of_binary_ops)
def binary_op(request):
    return request.param


list_of_unary_ops = [
    "negative",
    "abs",
    "absolute",
    "fabs",
    "sign",
    "conj",
    "exp",
    "exp2",
    "log",
    "log2",
    "log10",
    "expm1",
    "log1p",
    "sqrt",
    "square",
    "reciprocal",
    "conjugate",
    "floor",
    "ceil",
    "trunc",
]


@pytest.fixture(params=list_of_unary_ops)
def unary_op(request):
    return request.param


list_of_dtypes = [
    np.float32,
    np.float64,
]


@pytest.fixture(params=list_of_dtypes)
def input_arrays(request):
    # The size of input and out arrays to be used
    N = 2048
    a = np.array(np.random.random(N), request.param)
    b = np.array(np.random.random(N), request.param)
    return a, b


def test_binary_ops(filter_str, binary_op, input_arrays):
    if skip_test(filter_str):
        pytest.skip()

    a, b = input_arrays
    binop = getattr(np, binary_op)
    actual = np.empty(shape=a.shape, dtype=a.dtype)
    expected = np.empty(shape=a.shape, dtype=a.dtype)

    @njit
    def f(a, b):
        return binop(a, b)

    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), assert_auto_offloading():
        actual = f(a, b)

    expected = binop(a, b)
    np.testing.assert_allclose(actual, expected, rtol=1e-5, atol=0)


def test_unary_ops(filter_str, unary_op, input_arrays):
    if skip_test(filter_str):
        pytest.skip()

    # FIXME: Why does sign fail on Gen12 discrete graphics card?
    skip_ops = ["sign", "log", "log2", "log10", "expm1"]
    if unary_op in skip_ops and is_gen12(filter_str):
        pytest.skip()

    a = input_arrays[0]
    uop = getattr(np, unary_op)
    actual = np.empty(shape=a.shape, dtype=a.dtype)
    expected = np.empty(shape=a.shape, dtype=a.dtype)

    @njit
    def f(a):
        return uop(a)

    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), assert_auto_offloading():
        actual = f(a)

    expected = uop(a)
    np.testing.assert_allclose(actual, expected, rtol=1e-5, atol=0)



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numba.core import utils
from numba.core.cpu import CPUTargetOptions
from numba.core.descriptors import TargetDescriptor

from .target import DPPY_TARGET_NAME, DPPYTargetContext, DPPYTypingContext


class DPPYTarget(TargetDescriptor):
    options = CPUTargetOptions

    @utils.cached_property
    def _toplevel_target_context(self):
        # Lazily-initialized top-level target context, for all threads
        return DPPYTargetContext(self.typing_context, self._target_name)

    @utils.cached_property
    def _toplevel_typing_context(self):
        # Lazily-initialized top-level typing context, for all threads
        return DPPYTypingContext()

    @property
    def target_context(self):
        """
        The target context for DPPY targets.
        """
        return self._toplevel_target_context

    @property
    def typing_context(self):
        """
        The typing context for DPPY targets.
        """
        return self._toplevel_typing_context


# The global DPPY target
dppy_target = DPPYTarget(DPPY_TARGET_NAME)



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import ctypes
from inspect import signature
from types import FunctionType

import dpctl
import dpctl.program as dpctl_prog
import numpy as np
from numba.core import compiler, ir, types
from numba.core.compiler import CompilerBase, DefaultPassBuilder
from numba.core.compiler_lock import global_compiler_lock
from numba.core.typing.templates import AbstractTemplate, ConcreteTemplate

from numba_dppy import config
from numba_dppy.dppy_array_type import DPPYArray
from numba_dppy.dppy_parfor_diagnostics import ExtendedParforDiagnostics
from numba_dppy.driver import USMNdArrayType
from numba_dppy.utils import (
    as_usm_obj,
    assert_no_return,
    copy_from_numpy_to_usm_obj,
    copy_to_numpy_from_usm_obj,
    has_usm_memory,
)

from . import spirv_generator
from .dppy_passbuilder import DPPYPassBuilder

_NUMBA_DPPY_READ_ONLY = "read_only"
_NUMBA_DPPY_WRITE_ONLY = "write_only"
_NUMBA_DPPY_READ_WRITE = "read_write"


def _raise_no_device_found_error():
    error_message = (
        "No SYCL device specified. "
        "Usage : jit_fn[device, globalsize, localsize](...)"
    )
    raise ValueError(error_message)


def _raise_invalid_kernel_enqueue_args():
    error_message = (
        "Incorrect number of arguments for enquing dppy.kernel. "
        "Usage: device_env, global size, local size. "
        "The local size argument is optional."
    )
    raise ValueError(error_message)


def get_ordered_arg_access_types(pyfunc, access_types):
    # Construct a list of access type of each arg according to their position
    ordered_arg_access_types = []
    sig = signature(pyfunc, follow_wrapped=False)
    for idx, arg_name in enumerate(sig.parameters):
        if access_types:
            for key in access_types:
                if arg_name in access_types[key]:
                    ordered_arg_access_types.append(key)
        if len(ordered_arg_access_types) <= idx:
            ordered_arg_access_types.append(None)

    return ordered_arg_access_types


class DPPYCompiler(CompilerBase):
    """DPPY Compiler"""

    def define_pipelines(self):
        # this maintains the objmode fallback behaviour
        pms = []
        self.state.parfor_diagnostics = ExtendedParforDiagnostics()
        self.state.metadata["parfor_diagnostics"] = self.state.parfor_diagnostics
        if not self.state.flags.force_pyobject:
            # print("Numba-DPPY [INFO]: Using Numba-DPPY pipeline")
            pms.append(DPPYPassBuilder.define_nopython_pipeline(self.state))
        if self.state.status.can_fallback or self.state.flags.force_pyobject:
            pms.append(DefaultPassBuilder.define_objectmode_pipeline(self.state))
        return pms


@global_compiler_lock
def compile_with_dppy(pyfunc, return_type, args, is_kernel, debug=None):
    """
    Compiles with Numba_dppy's pipeline and returns the compiled result.

    Args:
        pyfunc: The Python function to be compiled.
        return_type: The Numba type of the return value.
        args: The list of arguments sent to the Python function.
        is_kernel (bool): Indicates whether the function is decorated
            with @dppy.kernel or not.
        debug (bool): Flag to turn debug mode ON/OFF.

    Returns:
        cres: Compiled result.

    Raises:
        TypeError: @dppy.kernel does not allow users to return any
            value. TypeError is raised when users do.

    """
    # First compilation will trigger the initialization of the OpenCL backend.
    from .descriptor import dppy_target

    typingctx = dppy_target.typing_context
    targetctx = dppy_target.target_context

    flags = compiler.Flags()
    # Do not compile (generate native code), just lower (to LLVM)
    flags.debuginfo = config.DEBUGINFO_DEFAULT
    flags.no_compile = True
    flags.no_cpython_wrapper = True
    flags.nrt = False

    if debug is not None:
        flags.debuginfo = debug

    # Run compilation pipeline
    if isinstance(pyfunc, FunctionType):
        cres = compiler.compile_extra(
            typingctx=typingctx,
            targetctx=targetctx,
            func=pyfunc,
            args=args,
            return_type=return_type,
            flags=flags,
            locals={},
            pipeline_class=DPPYCompiler,
        )
    elif isinstance(pyfunc, ir.FunctionIR):
        cres = compiler.compile_ir(
            typingctx=typingctx,
            targetctx=targetctx,
            func_ir=pyfunc,
            args=args,
            return_type=return_type,
            flags=flags,
            locals={},
            pipeline_class=DPPYCompiler,
        )
    else:
        assert 0

    if is_kernel:
        assert_no_return(cres.signature.return_type)

    # Linking depending libraries
    library = cres.library
    library.finalize()

    return cres


def compile_kernel(sycl_queue, pyfunc, args, access_types, debug=None):
    # For any array we only accept numba_dppy.dppy_array_type.DPPYArray
    for arg in args:
        if isinstance(arg, types.npytypes.Array) and not isinstance(arg, DPPYArray):
            raise TypeError(
                "We only accept DPPYArray as type of array-like objects. We received %s"
                % (type(arg))
            )

    if config.DEBUG:
        print("compile_kernel", args)
        debug = True
    if not sycl_queue:
        # We expect the sycl_queue to be provided when this function is called
        raise ValueError("SYCL queue is required for compiling a kernel")

    cres = compile_with_dppy(
        pyfunc=pyfunc, return_type=None, args=args, is_kernel=True, debug=debug
    )
    func = cres.library.get_function(cres.fndesc.llvm_func_name)
    kernel = cres.target_context.prepare_ocl_kernel(func, cres.signature.args)

    # A reference to the target context is stored in the DPPYKernel to
    # reference the context later in code generation. For example, we link
    # the kernel object with a spir_func defining atomic operations only
    # when atomic operations are used in the kernel.
    oclkern = DPPYKernel(
        context=cres.target_context,
        sycl_queue=sycl_queue,
        llvm_module=kernel.module,
        name=kernel.name,
        argtypes=cres.signature.args,
        ordered_arg_access_types=access_types,
    )
    return oclkern


def compile_kernel_parfor(sycl_queue, func_ir, args, args_with_addrspaces, debug=None):
    # For any array we only accept numba_dppy.dppy_array_type.DPPYArray
    for arg in args_with_addrspaces:
        if isinstance(arg, types.npytypes.Array) and not isinstance(arg, DPPYArray):
            raise TypeError(
                "We only accept DPPYArray as type of array-like objects. We received %s"
                % (type(arg))
            )
    if config.DEBUG:
        print("compile_kernel_parfor", args)
        for a in args_with_addrspaces:
            print(a, type(a))
            if isinstance(a, types.npytypes.Array):
                print("addrspace:", a.addrspace)

    cres = compile_with_dppy(
        pyfunc=func_ir,
        return_type=None,
        args=args_with_addrspaces,
        is_kernel=True,
        debug=debug,
    )
    func = cres.library.get_function(cres.fndesc.llvm_func_name)

    if config.DEBUG:
        print("compile_kernel_parfor signature", cres.signature.args)
        for a in cres.signature.args:
            print(a, type(a))

    kernel = cres.target_context.prepare_ocl_kernel(func, cres.signature.args)
    oclkern = DPPYKernel(
        context=cres.target_context,
        sycl_queue=sycl_queue,
        llvm_module=kernel.module,
        name=kernel.name,
        argtypes=args_with_addrspaces,
    )

    return oclkern


def compile_dppy_func(pyfunc, return_type, args, debug=None):
    cres = compile_with_dppy(
        pyfunc=pyfunc, return_type=return_type, args=args, is_kernel=False, debug=debug
    )
    func = cres.library.get_function(cres.fndesc.llvm_func_name)
    cres.target_context.mark_ocl_device(func)
    devfn = DPPYFunction(cres)

    class dppy_function_template(ConcreteTemplate):
        key = devfn
        cases = [cres.signature]

    cres.typing_context.insert_user_function(devfn, dppy_function_template)
    libs = [cres.library]
    cres.target_context.insert_user_function(devfn, cres.fndesc, libs)
    return devfn


# Compile dppy function template
def compile_dppy_func_template(pyfunc, debug=None):
    """Compile a DPPYFunctionTemplate"""
    from .descriptor import dppy_target

    dft = DPPYFunctionTemplate(pyfunc, debug=debug)

    class dppy_function_template(AbstractTemplate):
        key = dft

        def generic(self, args, kws):
            assert not kws
            return dft.compile(args)

    typingctx = dppy_target.typing_context
    typingctx.insert_user_function(dft, dppy_function_template)
    return dft


class DPPYFunctionTemplate(object):
    """Unmaterialized dppy function"""

    def __init__(self, pyfunc, debug=None):
        self.py_func = pyfunc
        self.debug = debug
        # self.inline = inline
        self._compileinfos = {}

    def compile(self, args):
        """Compile the function for the given argument types.

        Each signature is compiled once by caching the compiled function inside
        this object.
        """
        if args not in self._compileinfos:
            cres = compile_with_dppy(
                pyfunc=self.py_func,
                return_type=None,
                args=args,
                is_kernel=False,
                debug=self.debug,
            )
            func = cres.library.get_function(cres.fndesc.llvm_func_name)
            cres.target_context.mark_ocl_device(func)
            first_definition = not self._compileinfos
            self._compileinfos[args] = cres
            libs = [cres.library]

            if first_definition:
                # First definition
                cres.target_context.insert_user_function(self, cres.fndesc, libs)
            else:
                cres.target_context.add_user_function(self, cres.fndesc, libs)

        else:
            cres = self._compileinfos[args]

        return cres.signature


class DPPYFunction(object):
    def __init__(self, cres):
        self.cres = cres


def _ensure_valid_work_item_grid(val, sycl_queue):

    if not isinstance(val, (tuple, list, int)):
        error_message = "Cannot create work item dimension from provided argument"
        raise ValueError(error_message)

    if isinstance(val, int):
        val = [val]

    # TODO: we need some way to check the max dimensions
    """
    if len(val) > device_env.get_max_work_item_dims():
        error_message = ("Unsupported number of work item dimensions ")
        raise ValueError(error_message)
    """

    return list(
        val[::-1]
    )  # reversing due to sycl and opencl interop kernel range mismatch semantic


def _ensure_valid_work_group_size(val, work_item_grid):

    if not isinstance(val, (tuple, list, int)):
        error_message = "Cannot create work item dimension from provided argument"
        raise ValueError(error_message)

    if isinstance(val, int):
        val = [val]

    if len(val) != len(work_item_grid):
        error_message = (
            "Unsupported number of work item dimensions, "
            + "dimensions of global and local work items has to be the same "
        )
        raise ValueError(error_message)

    return list(
        val[::-1]
    )  # reversing due to sycl and opencl interop kernel range mismatch semantic


class DPPYKernelBase(object):
    """Define interface for configurable kernels"""

    def __init__(self):
        self.global_size = []
        self.local_size = []
        self.sycl_queue = None

        # list of supported access types, stored in dict for fast lookup
        self.valid_access_types = {
            _NUMBA_DPPY_READ_ONLY: _NUMBA_DPPY_READ_ONLY,
            _NUMBA_DPPY_WRITE_ONLY: _NUMBA_DPPY_WRITE_ONLY,
            _NUMBA_DPPY_READ_WRITE: _NUMBA_DPPY_READ_WRITE,
        }

    def copy(self):
        return copy.copy(self)

    def configure(self, sycl_queue, global_size, local_size=None):
        """Configure the OpenCL kernel. The local_size can be None"""
        clone = self.copy()
        clone.global_size = global_size
        clone.local_size = local_size
        clone.sycl_queue = sycl_queue

        return clone

    def __getitem__(self, args):
        """Mimick CUDA python's square-bracket notation for configuration.
        This assumes the argument to be:
            `global size, local size`
        """
        ls = None
        nargs = len(args)
        # Check if the kernel enquing arguments are sane
        if nargs < 1 or nargs > 2:
            _raise_invalid_kernel_enqueue_args

        sycl_queue = dpctl.get_current_queue()

        gs = _ensure_valid_work_item_grid(args[0], sycl_queue)
        # If the optional local size argument is provided
        if nargs == 2 and args[1] != []:
            ls = _ensure_valid_work_group_size(args[1], gs)

        return self.configure(sycl_queue, gs, ls)


class DPPYKernel(DPPYKernelBase):
    """
    A OCL kernel object
    """

    def __init__(
        self,
        context,
        sycl_queue,
        llvm_module,
        name,
        argtypes,
        ordered_arg_access_types=None,
    ):
        super(DPPYKernel, self).__init__()
        self._llvm_module = llvm_module
        self.assembly = self.binary = llvm_module.__str__()
        self.entry_name = name
        self.argument_types = tuple(argtypes)
        self.ordered_arg_access_types = ordered_arg_access_types
        self._argloc = []
        self.sycl_queue = sycl_queue
        self.context = context

        dpctl_create_program_from_spirv_flags = []
        # First-time compilation using SPIRV-Tools
        if config.DEBUG:
            with open("llvm_kernel.ll", "w") as f:
                f.write(self.binary)

        if config.DEBUG or config.OPT == 0:
            # if debug is ON we need to pass additional
            # flags to igc.
            dpctl_create_program_from_spirv_flags = ["-g", "-cl-opt-disable"]

        self.spirv_bc = spirv_generator.llvm_to_spirv(
            self.context, self.assembly, self._llvm_module.as_bitcode()
        )

        # create a program
        self.program = dpctl_prog.create_program_from_spirv(
            self.sycl_queue,
            self.spirv_bc,
            " ".join(dpctl_create_program_from_spirv_flags),
        )
        #  create a kernel
        self.kernel = self.program.get_sycl_kernel(self.entry_name)

    def __call__(self, *args):
        """
        Create a list of the kernel arguments by unpacking pyobject values
        into ctypes values.
        """

        kernelargs = []
        internal_device_arrs = []
        for ty, val, access_type in zip(
            self.argument_types, args, self.ordered_arg_access_types
        ):
            self._unpack_argument(
                ty,
                val,
                self.sycl_queue,
                kernelargs,
                internal_device_arrs,
                access_type,
            )

        self.sycl_queue.submit(
            self.kernel, kernelargs, self.global_size, self.local_size
        )
        self.sycl_queue.wait()

        for ty, val, i_dev_arr, access_type in zip(
            self.argument_types,
            args,
            internal_device_arrs,
            self.ordered_arg_access_types,
        ):
            self._pack_argument(ty, val, self.sycl_queue, i_dev_arr, access_type)

    def _pack_argument(self, ty, val, sycl_queue, device_arr, access_type):
        """
        Copy device data back to host
        """
        if device_arr and (
            access_type not in self.valid_access_types
            or access_type in self.valid_access_types
            and self.valid_access_types[access_type] != _NUMBA_DPPY_READ_ONLY
        ):
            # We copy the data back from usm allocated data
            # container to original data container.
            usm_mem, orig_ndarr, packed_ndarr, packed = device_arr
            copy_to_numpy_from_usm_obj(usm_mem, packed_ndarr)
            if packed:
                np.copyto(orig_ndarr, packed_ndarr)

    def _unpack_device_array_argument(
        self, size, itemsize, buf, shape, strides, ndim, kernelargs
    ):
        """
        Implements the unpacking logic for array arguments.

        Args:
            size: Total number of elements in the array.
            itemsize: Size in bytes of each element in the array.
            buf: The pointer to the memory.
            shape: The shape of the array.
            ndim: Number of dimension.
            kernelargs: Array where the arguments of the kernel is stored.
        """
        # meminfo
        kernelargs.append(ctypes.c_size_t(0))
        # parent
        kernelargs.append(ctypes.c_size_t(0))
        kernelargs.append(ctypes.c_longlong(size))
        kernelargs.append(ctypes.c_longlong(itemsize))
        kernelargs.append(buf)
        for ax in range(ndim):
            kernelargs.append(ctypes.c_longlong(shape[ax]))
        for ax in range(ndim):
            kernelargs.append(ctypes.c_longlong(strides[ax]))

    def _unpack_argument(
        self, ty, val, sycl_queue, kernelargs, device_arrs, access_type
    ):
        """
        Unpacks the arguments that are to be passed to the SYCL kernel from
        Numba types to Ctypes.

        Args:
            ty: The data types of the kernel argument defined as in instance of
                numba.types.
            val: The value of the kernel argument.
            sycl_queue (dpctl.SyclQueue): A ``dpctl.SyclQueue`` object. The
                queue object will be used whenever USM memory allocation is
                needed during unpacking of an numpy.ndarray argument.
            kernelargs (list): The list of kernel arguments into which the
                current kernel argument will be appended.
            device_arrs (list): A list of tuples that is used to store the
                triples corresponding to the USM memorry allocated for an
                ``numpy.ndarray`` argument, a wrapper ``ndarray`` created from
                the USM memory, and the original ``ndarray`` argument.
            access_type : The type of access for an array argument.

        Raises:
            NotImplementedError: If the type of argument is not yet supported,
                then a ``NotImplementedError`` is raised.

        """

        device_arrs.append(None)

        if isinstance(ty, USMNdArrayType):
            self._unpack_device_array_argument(
                val.size,
                val.dtype.itemsize,
                val.usm_data,
                val.shape,
                val.strides,
                val.ndim,
                kernelargs,
            )
        elif isinstance(ty, types.Array):
            packed_val = val
            usm_mem = has_usm_memory(val)
            if usm_mem is None:
                default_behavior = self.check_for_invalid_access_type(access_type)
                usm_mem = as_usm_obj(val, queue=sycl_queue, copy=False)

                orig_val = val
                packed = False
                if not val.flags.c_contiguous:
                    # If the numpy.ndarray is not C-contiguous
                    # we pack the strided array into a packed array.
                    # This allows us to treat the data from here on as C-contiguous.
                    # While packing we treat the data as C-contiguous.
                    # We store the reference of both (strided and packed)
                    # array and during unpacking we use numpy.copyto() to copy
                    # the data back from the packed temporary array to the
                    # original strided array.
                    packed_val = val.flatten(order="C")
                    packed = True

                if (
                    default_behavior
                    or self.valid_access_types[access_type] == _NUMBA_DPPY_READ_ONLY
                    or self.valid_access_types[access_type] == _NUMBA_DPPY_READ_WRITE
                ):
                    copy_from_numpy_to_usm_obj(usm_mem, packed_val)

                device_arrs[-1] = (usm_mem, orig_val, packed_val, packed)

            self._unpack_device_array_argument(
                packed_val.size,
                packed_val.dtype.itemsize,
                usm_mem,
                packed_val.shape,
                packed_val.strides,
                packed_val.ndim,
                kernelargs,
            )
        elif ty == types.int64:
            cval = ctypes.c_longlong(val)
            kernelargs.append(cval)
        elif ty == types.uint64:
            cval = ctypes.c_ulonglong(val)
            kernelargs.append(cval)
        elif ty == types.int32:
            cval = ctypes.c_int(val)
            kernelargs.append(cval)
        elif ty == types.uint32:
            cval = ctypes.c_uint(val)
            kernelargs.append(cval)
        elif ty == types.float64:
            cval = ctypes.c_double(val)
            kernelargs.append(cval)
        elif ty == types.float32:
            cval = ctypes.c_float(val)
            kernelargs.append(cval)
        elif ty == types.boolean:
            cval = ctypes.c_uint8(int(val))
            kernelargs.append(cval)
        elif ty == types.complex64:
            raise NotImplementedError(ty, val)
        elif ty == types.complex128:
            raise NotImplementedError(ty, val)
        else:
            raise NotImplementedError(ty, val)

    def check_for_invalid_access_type(self, access_type):
        if access_type not in self.valid_access_types:
            msg = (
                "[!] %s is not a valid access type. "
                "Supported access types are [" % (access_type)
            )
            for key in self.valid_access_types:
                msg += " %s |" % (key)

            msg = msg[:-1] + "]"
            if access_type is not None:
                print(msg)
            return True
        else:
            return False


class JitDPPYKernel(DPPYKernelBase):
    def __init__(self, func, debug, access_types):

        super(JitDPPYKernel, self).__init__()

        self.py_func = func
        self.definitions = {}
        self.debug = debug
        self.access_types = access_types

        from .descriptor import dppy_target

        self.typingctx = dppy_target.typing_context

    def _get_argtypes(self, *args):
        """
        Convenience function to get the type of each argument.
        """
        return tuple([self.typingctx.resolve_argument_type(a) for a in args])

    def __call__(self, *args, **kwargs):
        assert not kwargs, "Keyword Arguments are not supported"
        try:
            current_queue = dpctl.get_current_queue()
        except:
            _raise_no_device_found_error()

        argtypes = self._get_argtypes(*args)
        kernel = self.specialize(argtypes, current_queue)
        cfg = kernel.configure(self.sycl_queue, self.global_size, self.local_size)
        cfg(*args)

    def specialize(self, argtypes, queue):
        # We specialize for argtypes and queue. These two are used as key for
        # caching as well.
        assert queue is not None

        sycl_ctx = None
        kernel = None
        # we were previously using the _env_ptr of the device_env, the sycl_queue
        # should be sufficient to cache the compiled kernel for now, but we should
        # use the device type to cache such kernels.
        key_definitions = argtypes
        result = self.definitions.get(key_definitions)
        if result:
            sycl_ctx, kernel = result

        if sycl_ctx and sycl_ctx == queue.sycl_context:
            return kernel
        else:
            kernel = compile_kernel(
                queue, self.py_func, argtypes, self.access_types, self.debug
            )
            self.definitions[key_definitions] = (queue.sycl_context, kernel)
        return kernel



#! /usr/bin/env python
# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dpctl
import numpy as np
import pytest
from numba import float32, float64, int32, int64, njit, vectorize

import numba_dppy as dppy
from numba_dppy.tests._helper import assert_auto_offloading

from . import _helper

list_of_filter_strs = [
    "opencl:gpu:0",
    "level_zero:gpu:0",
    "opencl:cpu:0",
]


@pytest.fixture(params=list_of_filter_strs)
def filter_str(request):
    return request.param


list_of_shape = [
    (100, 100),
    (100, (10, 10)),
    (100, (2, 5, 10)),
]


@pytest.fixture(params=list_of_shape)
def shape(request):
    return request.param


def test_njit(filter_str):
    if _helper.platform_not_supported(filter_str):
        pytest.skip()

    if _helper.skip_test(filter_str):
        pytest.skip()

    @vectorize(nopython=True)
    def axy(a, x, y):
        return a * x + y

    def f(a0, a1):
        return np.cos(axy(a0, np.sin(a1) - 1.0, 1.0))

    A = np.random.random(10)
    B = np.random.random(10)

    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), assert_auto_offloading():
        f_njit = njit(f)
        expected = f_njit(A, B)
        actual = f(A, B)

        max_abs_err = expected.sum() - actual.sum()
        assert max_abs_err < 1e-5


list_of_dtype = [
    (np.int32, int32),
    (np.float32, float32),
    (np.int64, int64),
    (np.float64, float64),
]


@pytest.fixture(params=list_of_dtype)
def dtypes(request):
    return request.param


list_of_input_type = ["array", "scalar"]


@pytest.fixture(params=list_of_input_type)
def input_type(request):
    return request.param


def test_vectorize(filter_str, shape, dtypes, input_type):
    if _helper.platform_not_supported(filter_str):
        pytest.skip()

    if _helper.skip_test(filter_str):
        pytest.skip()

    def vector_add(a, b):
        return a + b

    dtype, sig_dtype = dtypes
    sig = [sig_dtype(sig_dtype, sig_dtype)]
    size, shape = shape

    if input_type == "array":
        A = np.arange(size, dtype=dtype).reshape(shape)
        B = np.arange(size, dtype=dtype).reshape(shape)
    elif input_type == "scalar":
        A = dtype(1.2)
        B = dtype(2.3)

    with dpctl.device_context(filter_str):
        f = vectorize(sig, target="dppy")(vector_add)
        expected = f(A, B)
        actual = vector_add(A, B)

        max_abs_err = np.sum(expected) - np.sum(actual)
        assert max_abs_err < 1e-5
